#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0x7d5d22f9

# Compiled with Coconut version 1.3.1-post_dev18 [Dead Parrot]

# Coconut Header: -------------------------------------------------------------

import sys as _coconut_sys, os.path as _coconut_os_path
_coconut_file_path = _coconut_os_path.dirname(_coconut_os_path.abspath(__file__))
_coconut_sys.path.insert(0, _coconut_file_path)
from __coconut__ import _coconut, _coconut_NamedTuple, _coconut_MatchError, _coconut_tail_call, _coconut_tco, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_pipe, _coconut_star_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial
from __coconut__ import *
_coconut_sys.path.remove(_coconut_file_path)

# Compiled Coconut: -----------------------------------------------------------

# Copyright Arne Bachmann
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import builtins  # line 4
import codecs  # line 4
import enum  # line 4
import json  # line 4
import logging  # line 4
import os  # line 4
import shutil  # line 4
sys = _coconut_sys  # line 4
import time  # line 4
import traceback  # line 4
import unittest  # line 4
import uuid  # line 4
from io import BytesIO  # line 5
from io import BufferedRandom  # line 5
from io import TextIOWrapper  # line 5
try:  # Python 3  # line 6
    from unittest import mock  # Python 3  # line 6
except:  # installed via pip  # line 7
    import mock  # installed via pip  # line 7

testFolder = os.path.abspath(os.path.join(os.getcwd(), "test"))  # line 9
import configr  # optional dependency  # line 10
os.environ["TEST"] = testFolder  # needed to mock configr library calls in sos  # line 11
import sos  # import of package, not file  # line 12
sos.defaults["defaultbranch"] = "trunk"  # because sos.main() is never called  # line 13
sos.defaults["useChangesCommand"] = True  # because sos.main() is never called  # line 14

def sync():  # line 16
    if sys.version_info[:2] >= (3, 3):  # line 17
        os.sync()  # line 17


def determineFilesystemTimeResolution() -> 'float':  # line 20
    name = str(uuid.uuid4())  # type: str  # line 21
    with open(name, "w") as fd:  # create temporary file  # line 22
        fd.write("x")  # create temporary file  # line 22
    mt = os.stat(name).st_mtime  # type: float  # get current timestamp  # line 23
    while os.stat(name).st_mtime == mt:  # wait until timestamp modified  # line 24
        time.sleep(0.05)  # to avoid 0.00s bugs (came up some time for unknown reasons)  # line 25
        with open(name, "w") as fd:  # line 26
            fd.write("x")  # line 26
    mt, start, _count = os.stat(name).st_mtime, time.time(), 0  # line 27
    while os.stat(name).st_mtime == mt:  # now cound and measure time until modified again  # line 28
        time.sleep(0.05)  # line 29
        _count += 1  # line 30
        with open(name, "w") as fd:  # line 31
            fd.write("x")  # line 31
    os.unlink(name)  # line 32
    fsprecision = round(time.time() - start, 2)  # type: float  # line 33
    print("File system timestamp precision is %.2fs; wrote to the file %d times during that time" % (fsprecision, _count))  # line 34
    return fsprecision  # line 35


FS_PRECISION = determineFilesystemTimeResolution() * 1.05  # line 38


@_coconut_tco  # line 41
def debugTestRunner(post_mortem=None):  # line 41
    ''' Unittest runner doing post mortem debugging on failing tests. '''  # line 42
    import pdb  # line 43
    if post_mortem is None:  # line 44
        post_mortem = pdb.post_mortem  # line 44
    class DebugTestResult(unittest.TextTestResult):  # line 45
        def addError(_, test, err):  # called before tearDown()  # line 46
            traceback.print_exception(*err)  # line 47
            post_mortem(err[2])  # line 48
            super(DebugTestResult, _).addError(test, err)  # line 49
        def addFailure(_, test, err):  # line 50
            traceback.print_exception(*err)  # line 51
            post_mortem(err[2])  # line 52
            super(DebugTestResult, _).addFailure(test, err)  # line 53
    return _coconut_tail_call(unittest.TextTestRunner, resultclass=DebugTestResult)  # line 54

@_coconut_tco  # line 56
def wrapChannels(func: '_coconut.typing.Callable[..., Any]'):  # line 56
    ''' Wrap function call to capture and return strings emitted on stdout and stderr. '''  # line 57
    oldv = sys.argv  # line 58
    buf = TextIOWrapper(BufferedRandom(BytesIO(b"")), encoding=sos.UTF8)  # line 59
    sys.stdout = sys.stderr = buf  # line 60
    handler = logging.StreamHandler(buf)  # line 61
    logging.getLogger().addHandler(handler)  # line 62
    try:  # capture output into buf  # line 63
        func()  # capture output into buf  # line 63
    except Exception as E:  # line 64
        buf.write(str(E) + "\n")  # line 64
        traceback.print_exc(file=buf)  # line 64
    except SystemExit as F:  # line 65
        buf.write("EXIT CODE %s" % F.code + "\n")  # line 65
        traceback.print_exc(file=buf)  # line 65
    logging.getLogger().removeHandler(handler)  # line 66
    sys.argv, sys.stdout, sys.stderr = oldv, sys.__stdout__, sys.__stderr__  # TODO when run using pythonw.exe and/or no console, these could be None  # line 67
    buf.seek(0)  # line 68
    return _coconut_tail_call(buf.read)  # line 69

def mockInput(datas: '_coconut.typing.Sequence[str]', func) -> 'Any':  # line 71
    with mock.patch("builtins.input", side_effect=datas):  # line 72
        return func()  # line 72

def setRepoFlag(name: 'str', value: 'bool'):  # line 74
    with open(sos.metaFolder + os.sep + sos.metaFile, "r") as fd:  # line 75
        flags, branches, config = json.loads(fd.read())  # line 75
    flags[name] = value  # line 76
    with open(sos.metaFolder + os.sep + sos.metaFile, "w") as fd:  # line 77
        fd.write(json.dumps((flags, branches, config)))  # line 77

def checkRepoFlag(name: 'str', flag: '_coconut.typing.Optional[bool]'=None, value: '_coconut.typing.Optional[Any]'=None) -> 'bool':  # line 79
    with open(sos.metaFolder + os.sep + sos.metaFile, "r") as fd:  # line 80
        flags, branches, config = json.loads(fd.read())  # line 80
    return (name in flags and flags[name] == flag) if flag is not None else (name in config and config[name] == value)  # line 81


class Tests(unittest.TestCase):  # line 84
    ''' Entire test suite. '''  # line 85

    def setUp(_):  # line 87
        for entry in os.listdir(testFolder):  # cannot remove testFolder on Windows when using TortoiseSVN as VCS  # line 88
            resource = os.path.join(testFolder, entry)  # line 89
            shutil.rmtree(resource) if os.path.isdir(resource) else os.unlink(resource)  # line 90
        os.chdir(testFolder)  # line 91


    def assertAllIn(_, what: '_coconut.typing.Sequence[str]', where: 'Union[str, List[str]]', only: 'bool'=False):  # line 94
        for w in what:  # line 95
            _.assertIn(w, where)  # line 95
        if only:  # line 96
            _.assertEqual(len(what), len(where))  # line 96

    def assertAllNotIn(_, what: '_coconut.typing.Sequence[str]', where: 'Union[str, List[str]]'):  # line 98
        for w in what:  # line 99
            _.assertNotIn(w, where)  # line 99

    def assertInAll(_, what: 'str', where: '_coconut.typing.Sequence[str]'):  # line 101
        for w in where:  # line 102
            _.assertIn(what, w)  # line 102

    def assertInAny(_, what: 'str', where: '_coconut.typing.Sequence[str]'):  # line 104
        _.assertTrue(any((what in w for w in where)))  # line 104

    def assertNotInAny(_, what: 'str', where: '_coconut.typing.Sequence[str]'):  # line 106
        _.assertFalse(any((what in w for w in where)))  # line 106

    def createFile(_, number: 'Union[int, str]', contents: 'str'="x" * 10, prefix: '_coconut.typing.Optional[str]'=None):  # line 108
        if prefix and not os.path.exists(prefix):  # line 109
            os.makedirs(prefix)  # line 109
        with open(("." if prefix is None else prefix) + os.sep + (("file%d" % number) if isinstance(number, int) else number), "wb") as fd:  # line 110
            fd.write(contents if isinstance(contents, bytes) else contents.encode("cp1252"))  # line 110

    def existsFile(_, number: 'Union[int, str]', expectedContents: 'bytes'=None) -> 'bool':  # line 112
        if not os.path.exists(("." + os.sep + "file%d" % number) if isinstance(number, int) else number):  # line 113
            return False  # line 113
        if expectedContents is None:  # line 114
            return True  # line 114
        with open(("." + os.sep + "file%d" % number) if isinstance(number, int) else number, "rb") as fd:  # line 115
            return fd.read() == expectedContents  # line 115

    def testAccessor(_):  # line 117
        a = sos.Accessor({"a": 1})  # line 118
        _.assertEqual((1, 1), (a["a"], a.a))  # line 119

    def testGetAnyOfmap(_):  # line 121
        _.assertEqual(2, sos.getAnyOfMap({"a": 1, "b": 2}, ["x", "b"]))  # line 122
        _.assertIsNone(sos.getAnyOfMap({"a": 1, "b": 2}, []))  # line 123

    def testAjoin(_):  # line 125
        _.assertEqual("a1a2", sos.ajoin("a", ["1", "2"]))  # line 126
        _.assertEqual("* a\n* b", sos.ajoin("* ", ["a", "b"], "\n"))  # line 127

    def testFindChanges(_):  # line 129
        m = sos.Metadata(os.getcwd())  # line 130
        try:  # line 131
            sos.config(["set", "texttype", "*"])  # line 131
        except SystemExit as E:  # line 132
            _.assertEqual(0, E.code)  # line 132
        try:  # will be stripped from leading paths anyway  # line 133
            sos.config(["set", "ignores", "test/*.cfg;D:\\apps\\*.cfg.bak"])  # will be stripped from leading paths anyway  # line 133
        except SystemExit as E:  # line 134
            _.assertEqual(0, E.code)  # line 134
        m = sos.Metadata(os.getcwd())  # reload from file system  # line 135
        for file in [f for f in os.listdir() if f.endswith(".bak")]:  # remove configuration file  # line 136
            os.unlink(file)  # remove configuration file  # line 136
        _.createFile(1, "1")  # line 137
        m.createBranch(0)  # line 138
        _.assertEqual(1, len(m.paths))  # line 139
        time.sleep(FS_PRECISION)  # time required by filesystem time resolution issues  # line 140
        _.createFile(1, "2")  # modify existing file  # line 141
        _.createFile(2, "2")  # add another file  # line 142
        m.loadCommit(0, 0)  # line 143
        changes, msg = m.findChanges()  # detect time skew  # line 144
        _.assertEqual(1, len(changes.additions))  # line 145
        _.assertEqual(0, len(changes.deletions))  # line 146
        _.assertEqual(1, len(changes.modifications))  # line 147
        _.assertEqual(0, len(changes.moves))  # line 148
        m.paths.update(changes.additions)  # line 149
        m.paths.update(changes.modifications)  # line 150
        _.createFile(2, "12")  # modify file again  # line 151
        changes, msg = m.findChanges(0, 1)  # by size, creating new commit  # line 152
        _.assertEqual(0, len(changes.additions))  # line 153
        _.assertEqual(0, len(changes.deletions))  # line 154
        _.assertEqual(1, len(changes.modifications))  # line 155
        _.assertEqual(0, len(changes.moves))  # line 156
        _.assertTrue(os.path.exists(sos.revisionFolder(0, 1)))  # line 157
        _.assertTrue(os.path.exists(sos.revisionFolder(0, 1, file="03b69bc801ae11f1ff2a71a50f165996d0ad681b4f822df13329a27e53f0fcd2")))  # line 158
# TODO test moves

    def testPatternPaths(_):  # line 161
        sos.offline(options=["--track"])  # line 162
        os.mkdir("sub")  # line 163
        _.createFile("sub" + os.sep + "file1", "sdfsdf")  # line 164
        sos.add("sub", "sub/file?")  # line 165
        sos.commit("test")  # should pick up sub/file1 pattern  # line 166
        _.assertEqual(2, len(os.listdir(sos.revisionFolder(0, 1))))  # sub/file1 was added  # line 167
        _.createFile(1)  # line 168
        try:  # should not commit anything, as the file in base folder doesn't match the tracked pattern  # line 169
            sos.commit("nothing")  # should not commit anything, as the file in base folder doesn't match the tracked pattern  # line 169
            _.fail()  # should not commit anything, as the file in base folder doesn't match the tracked pattern  # line 169
        except:  # line 170
            pass  # line 170

    def testNoArgs(_):  # line 172
        pass  # call "sos" without arguments should simply show help or info about missing arguments  # line 173

    def testAutoUpgrade(_):  # line 175
        sos.offline()  # line 176
        with codecs.open(sos.encode(os.path.join(sos.metaFolder, sos.metaFile)), "r", encoding=sos.UTF8) as fd:  # line 177
            repo, branches, config = json.load(fd)  # line 177
        repo["version"] = "0"  # lower than any pip version  # line 178
        branches[:] = [branch[:5] for branch in branches]  # simulate some older state  # line 179
        del repo["format"]  # simulate pre-1.3.5  # line 180
        with codecs.open(sos.encode(os.path.join(sos.metaFolder, sos.metaFile)), "w", encoding=sos.UTF8) as fd:  # line 181
            json.dump((repo, branches, config), fd, ensure_ascii=False)  # line 181
        out = wrapChannels(lambda: sos.status(options=["--repo"]))  # line 182
        _.assertAllIn(["Upgraded repository metadata to match SOS version '2018.1210.3028'", "Upgraded repository metadata to match SOS version '1.3.5'"], out)  # line 183

    def testFastBranching(_):  # line 185
        _.createFile(1)  # line 186
        sos.offline(options=["--strict"])  # b0/r0 = ./file1  # line 187
        _.createFile(2)  # line 188
        os.unlink("file1")  # line 189
        sos.commit()  # b0/r1 = ./file2  # line 190
        sos.branch(options=["--fast", "--last"])  # branch b1 from b0/1 TODO modify once --fast becomes the new normal  # line 191
        _.assertAllIn([sos.metaFile, sos.metaBack, "b0", "b1"], os.listdir(sos.metaFolder), only=True)  # line 192
        _.createFile(3)  # line 193
        sos.commit()  # b1/r2 = ./file2, ./file3  # line 194
        _.assertAllIn([sos.metaFile, "r2"], os.listdir(sos.branchFolder(1)), only=True)  # line 195
        sos.branch(options=["--fast", "--last"])  # branch b2 from b1/2  # line 196
        sos.destroy("0")  # remove parent of b1 and transitive parent of b2  # line 197
        _.assertAllIn([sos.metaFile, sos.metaBack, "b0_last", "b1", "b2"], os.listdir(sos.metaFolder), only=True)  # branch 0 was removed  # line 198
        _.assertAllIn([sos.metaFile, "r0", "r1", "r2"], os.listdir(sos.branchFolder(1)), only=True)  # revisions were copied to branch 1  # line 199
        _.assertAllIn([sos.metaFile, "r0", "r1", "r2"], os.listdir(sos.branchFolder(2)), only=True)  # revisions were copied to branch 1  # line 200
# TODO test also other functions like status --repo, log

    def testGetParentBranch(_):  # line 203
        m = sos.Accessor({"branches": {0: sos.Accessor({"parent": None, "revision": None}), 1: sos.Accessor({"parent": 0, "revision": 1})}})  # line 204
        _.assertEqual(0, sos.Metadata.getParentBranch(m, 1, 0))  # line 205
        _.assertEqual(0, sos.Metadata.getParentBranch(m, 1, 1))  # line 206
        _.assertEqual(1, sos.Metadata.getParentBranch(m, 1, 2))  # line 207
        _.assertEqual(0, sos.Metadata.getParentBranch(m, 0, 10))  # line 208

    def testTokenizeGlobPattern(_):  # line 210
        _.assertEqual([], sos.tokenizeGlobPattern(""))  # line 211
        _.assertEqual([sos.GlobBlock(False, "*", 0)], sos.tokenizeGlobPattern("*"))  # line 212
        _.assertEqual([sos.GlobBlock(False, "*", 0), sos.GlobBlock(False, "???", 1)], sos.tokenizeGlobPattern("*???"))  # line 213
        _.assertEqual([sos.GlobBlock(True, "x", 0), sos.GlobBlock(False, "*", 1), sos.GlobBlock(True, "x", 2)], sos.tokenizeGlobPattern("x*x"))  # line 214
        _.assertEqual([sos.GlobBlock(True, "x", 0), sos.GlobBlock(False, "*", 1), sos.GlobBlock(False, "??", 2), sos.GlobBlock(False, "*", 4), sos.GlobBlock(True, "x", 5)], sos.tokenizeGlobPattern("x*??*x"))  # line 215
        _.assertEqual([sos.GlobBlock(False, "?", 0), sos.GlobBlock(True, "abc", 1), sos.GlobBlock(False, "*", 4)], sos.tokenizeGlobPattern("?abc*"))  # line 216

    def testTokenizeGlobPatterns(_):  # line 218
        try:  # because number of literal strings differs  # line 219
            sos.tokenizeGlobPatterns("x*x", "x*")  # because number of literal strings differs  # line 219
            _.fail()  # because number of literal strings differs  # line 219
        except:  # line 220
            pass  # line 220
        try:  # because glob patterns differ  # line 221
            sos.tokenizeGlobPatterns("x*", "x?")  # because glob patterns differ  # line 221
            _.fail()  # because glob patterns differ  # line 221
        except:  # line 222
            pass  # line 222
        try:  # glob patterns differ, regardless of position  # line 223
            sos.tokenizeGlobPatterns("x*", "?x")  # glob patterns differ, regardless of position  # line 223
            _.fail()  # glob patterns differ, regardless of position  # line 223
        except:  # line 224
            pass  # line 224
        sos.tokenizeGlobPatterns("x*", "*x")  # succeeds, because glob patterns match (differ only in position)  # line 225
        sos.tokenizeGlobPatterns("*xb?c", "*x?bc")  # succeeds, because glob patterns match (differ only in position)  # line 226
        try:  # succeeds, because glob patterns match (differ only in position)  # line 227
            sos.tokenizeGlobPatterns("a???b*", "ab???*")  # succeeds, because glob patterns match (differ only in position)  # line 227
            _.fail()  # succeeds, because glob patterns match (differ only in position)  # line 227
        except:  # line 228
            pass  # line 228

    def testConvertGlobFiles(_):  # line 230
        _.assertEqual(["xxayb", "aacb"], [r[1] for r in sos.convertGlobFiles(["axxby", "aabc"], *sos.tokenizeGlobPatterns("a*b?", "*a?b"))])  # line 231
        _.assertEqual(["1qq2ww3", "1abcbx2xbabc3"], [r[1] for r in sos.convertGlobFiles(["qqxbww", "abcbxxbxbabc"], *sos.tokenizeGlobPatterns("*xb*", "1*2*3"))])  # line 232

    def testFolderRemove(_):  # line 234
        m = sos.Metadata(os.getcwd())  # line 235
        _.createFile(1)  # line 236
        _.createFile("a", prefix="sub")  # line 237
        sos.offline()  # line 238
        _.createFile(2)  # line 239
        os.unlink("sub" + os.sep + "a")  # line 240
        os.rmdir("sub")  # line 241
        changes = sos.changes()  # TODO replace by output check  # line 242
        _.assertEqual(1, len(changes.additions))  # line 243
        _.assertEqual(0, len(changes.modifications))  # line 244
        _.assertEqual(1, len(changes.deletions))  # line 245
        _.createFile("a", prefix="sub")  # line 246
        changes = sos.changes()  # line 247
        _.assertEqual(0, len(changes.deletions))  # line 248

    def testSwitchConflict(_):  # line 250
        sos.offline(options=["--strict"])  # (r0)  # line 251
        _.createFile(1)  # line 252
        sos.commit()  # add file (r1)  # line 253
        os.unlink("file1")  # line 254
        sos.commit()  # remove (r2)  # line 255
        _.createFile(1, "something else")  # line 256
        sos.commit()  # (r3)  # line 257
        sos.switch("/1")  # updates file1 - marked as MOD, because mtime was changed  # line 258
        _.existsFile(1, "x" * 10)  # line 259
        sos.switch("/2", ["--force"])  # remove file1 requires --force, because size/content (or mtime in non-strict mode) is different to head of branch  # line 260
        sos.switch("/0")  # do nothing, as file1 is already removed  # line 261
        sos.switch("/1")  # add file1 back  # line 262
        sos.switch("/", ["--force"])  # requires force because changed vs. head of branch  # line 263
        _.existsFile(1, "something else")  # line 264

    def testComputeSequentialPathSet(_):  # line 266
        os.makedirs(sos.revisionFolder(0, 0))  # line 267
        os.makedirs(sos.revisionFolder(0, 1))  # line 268
        os.makedirs(sos.revisionFolder(0, 2))  # line 269
        os.makedirs(sos.revisionFolder(0, 3))  # line 270
        os.makedirs(sos.revisionFolder(0, 4))  # line 271
        m = sos.Metadata(os.getcwd())  # line 272
        m.branch = 0  # line 273
        m.commit = 2  # line 274
        m.saveBranches()  # line 275
        m.paths = {"./a": sos.PathInfo("", 0, 0, "")}  # line 276
        m.saveCommit(0, 0)  # initial  # line 277
        m.paths["./a"] = sos.PathInfo("", 1, 0, "")  # line 278
        m.saveCommit(0, 1)  # mod  # line 279
        m.paths["./b"] = sos.PathInfo("", 0, 0, "")  # line 280
        m.saveCommit(0, 2)  # add  # line 281
        m.paths["./a"] = sos.PathInfo("", None, 0, "")  # line 282
        m.saveCommit(0, 3)  # del  # line 283
        m.paths["./a"] = sos.PathInfo("", 2, 0, "")  # line 284
        m.saveCommit(0, 4)  # readd  # line 285
        m.commits = {i: sos.CommitInfo(i, 0, None) for i in range(5)}  # line 286
        m.saveBranch(0)  # line 287
        m.branches = {0: sos.BranchInfo(0, 0), 1: sos.BranchInfo(1, 0)}  # line 288
        m.saveBranches()  # line 289
        m.computeSequentialPathSet(0, 4)  # line 290
        _.assertEqual(2, len(m.paths))  # line 291

    def testParseRevisionString(_):  # line 293
        m = sos.Metadata(os.getcwd())  # line 294
        m.branch = 1  # line 295
        m.commits = {0: 0, 1: 1, 2: 2}  # line 296
        _.assertEqual((1, 3), m.parseRevisionString("3"))  # line 297
        _.assertEqual((2, 3), m.parseRevisionString("2/3"))  # line 298
        _.assertEqual((1, -1), m.parseRevisionString(None))  # line 299
        _.assertEqual((1, -1), m.parseRevisionString(""))  # line 300
        _.assertEqual((2, -1), m.parseRevisionString("2/"))  # line 301
        _.assertEqual((1, -2), m.parseRevisionString("/-2"))  # line 302
        _.assertEqual((1, -1), m.parseRevisionString("/"))  # line 303

    def testOfflineEmpty(_):  # line 305
        os.mkdir("." + os.sep + sos.metaFolder)  # line 306
        try:  # line 307
            sos.offline("trunk")  # line 307
            _.fail()  # line 307
        except SystemExit as E:  # line 308
            _.assertEqual(1, E.code)  # line 308
        os.rmdir("." + os.sep + sos.metaFolder)  # line 309
        sos.offline("test")  # line 310
        _.assertIn(sos.metaFolder, os.listdir("."))  # line 311
        _.assertAllIn(["b0", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder))  # line 312
        _.assertAllIn(["r0", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0"))  # line 313
        _.assertEqual(2, len(os.listdir("." + os.sep + sos.metaFolder)))  # only branch folder and meta data file  # line 314
        _.assertEqual(2, len(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0")))  # only commit folder and meta data file  # line 315
        _.assertEqual(1, len(os.listdir(sos.revisionFolder(0, 0))))  # only meta data file  # line 316

    def testOfflineWithFiles(_):  # line 318
        _.createFile(1, "x" * 100)  # line 319
        _.createFile(2)  # line 320
        sos.offline("test")  # line 321
        _.assertAllIn(["file1", "file2", sos.metaFolder], os.listdir("."))  # line 322
        _.assertAllIn(["b0", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder))  # line 323
        _.assertAllIn(["r0", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0"))  # line 324
        _.assertAllIn([sos.metaFile, "03b69bc801ae11f1ff2a71a50f165996d0ad681b4f822df13329a27e53f0fcd2", "b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa"], os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0" + os.sep + "r0"))  # line 325
        _.assertEqual(2, len(os.listdir("." + os.sep + sos.metaFolder)))  # only branch folder and meta data file  # line 326
        _.assertEqual(2, len(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0")))  # only commit folder and meta data file  # line 327
        _.assertEqual(3, len(os.listdir(sos.revisionFolder(0, 0))))  # only meta data file plus branch base file copies  # line 328

    def testBranch(_):  # line 330
        _.createFile(1, "x" * 100)  # line 331
        _.createFile(2)  # line 332
        sos.offline("test")  # b0/r0  # line 333
        sos.branch("other")  # b1/r0  # line 334
        _.assertAllIn(["b0", "b1", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder))  # line 335
        _.assertEqual(list(sorted(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0"))), list(sorted(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b1"))))  # line 336
        _.assertEqual(list(sorted(os.listdir(sos.revisionFolder(0, 0)))), list(sorted(os.listdir(sos.revisionFolder(1, 0)))))  # line 338
        _.createFile(1, "z")  # modify file  # line 340
        sos.branch()  # b2/r0  branch to unnamed branch with modified file tree contents  # line 341
        _.assertNotEqual(os.stat("." + os.sep + sos.metaFolder + os.sep + "b1" + os.sep + "r0" + os.sep + "b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa").st_size, os.stat("." + os.sep + sos.metaFolder + os.sep + "b2" + os.sep + "r0" + os.sep + "b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa").st_size)  # line 342
        _.createFile(3, "z")  # line 344
        sos.branch("from_last_revision", options=["--last", "--stay"])  # b3/r0 create copy of other file1,file2 and don't switch  # line 345
        _.assertEqual(list(sorted(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b3" + os.sep + "r0"))), list(sorted(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b2" + os.sep + "r0"))))  # line 346
# Check sos.status output which branch is marked


    def testComittingAndChanges(_):  # line 351
        _.createFile(1, "x" * 100)  # line 352
        _.createFile(2)  # line 353
        sos.offline("test")  # line 354
        changes = sos.changes()  # line 355
        _.assertEqual(0, len(changes.additions))  # line 356
        _.assertEqual(0, len(changes.deletions))  # line 357
        _.assertEqual(0, len(changes.modifications))  # line 358
        _.createFile(1, "z")  # size change  # line 359
        changes = sos.changes()  # line 360
        _.assertEqual(0, len(changes.additions))  # line 361
        _.assertEqual(0, len(changes.deletions))  # line 362
        _.assertEqual(1, len(changes.modifications))  # line 363
        sos.commit("message")  # line 364
        _.assertAllIn(["r0", "r1", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0"))  # line 365
        _.assertAllIn([sos.metaFile, "b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa"], os.listdir(sos.revisionFolder(0, 1)))  # line 366
        _.assertEqual(2, len(os.listdir(sos.revisionFolder(0, 1))))  # no further files, only the modified one  # line 367
        _.assertEqual(1, len(sos.changes("/0").modifications))  # vs. explicit revision on current branch  # line 368
        _.assertEqual(1, len(sos.changes("0/0").modifications))  # vs. explicit branch/revision  # line 369
        _.createFile(1, "")  # modify to empty file, mentioned in meta data, but not stored as own file  # line 370
        os.unlink("file2")  # line 371
        changes = sos.changes()  # line 372
        _.assertEqual(0, len(changes.additions))  # line 373
        _.assertEqual(1, len(changes.deletions))  # line 374
        _.assertEqual(1, len(changes.modifications))  # line 375
        sos.commit("modified")  # line 376
        _.assertEqual(1, len(os.listdir(sos.revisionFolder(0, 2))))  # no additional files, only mentions in metadata  # line 377
        try:  # expecting Exit due to no changes  # line 378
            sos.commit("nothing")  # expecting Exit due to no changes  # line 378
            _.fail()  # expecting Exit due to no changes  # line 378
        except:  # line 379
            pass  # line 379

    def testGetBranch(_):  # line 381
        m = sos.Metadata(os.getcwd())  # line 382
        m.branch = 1  # current branch  # line 383
        m.branches = {0: sos.BranchInfo(0, 0, "trunk")}  # line 384
        _.assertEqual(27, m.getBranchByName(27))  # line 385
        _.assertEqual(0, m.getBranchByName("trunk"))  # line 386
        _.assertEqual(1, m.getBranchByName(""))  # split from "/"  # line 387
        _.assertIsNone(m.getBranchByName("unknown"))  # line 388
        m.commits = {0: sos.CommitInfo(0, 0, "bla")}  # line 389
        _.assertEqual(13, m.getRevisionByName("13"))  # line 390
        _.assertEqual(0, m.getRevisionByName("bla"))  # line 391
        _.assertEqual(-1, m.getRevisionByName(""))  # split from "/"  # line 392

    def testTagging(_):  # line 394
        m = sos.Metadata(os.getcwd())  # line 395
        sos.offline()  # line 396
        _.createFile(111)  # line 397
        sos.commit("tag", ["--tag"])  # line 398
        out = wrapChannels(lambda: sos.log()).replace("\r", "").split("\n")  # line 399
        _.assertTrue(any(("|tag" in line and line.endswith("|TAG") for line in out)))  # line 400
        _.createFile(2)  # line 401
        try:  # line 402
            sos.commit("tag")  # line 402
            _.fail()  # line 402
        except:  # line 403
            pass  # line 403
        sos.commit("tag-2", ["--tag"])  # line 404
        out = wrapChannels(lambda: sos.ls(options=["--tags"])).replace("\r", "")  # line 405
        _.assertIn("TAG tag", out)  # line 406

    def testSwitch(_):  # line 408
        _.createFile(1, "x" * 100)  # line 409
        _.createFile(2, "y")  # line 410
        sos.offline("test")  # file1-2  in initial branch commit  # line 411
        sos.branch("second")  # file1-2  switch, having same files  # line 412
        sos.switch("0")  # no change  switch back, no problem  # line 413
        sos.switch("second")  # no change  # switch back, no problem  # line 414
        _.createFile(3, "y")  # generate a file  # line 415
        try:  # uncommited changes detected  # line 416
            sos.switch("test")  # uncommited changes detected  # line 416
            _.fail()  # uncommited changes detected  # line 416
        except SystemExit as E:  # line 417
            _.assertEqual(1, E.code)  # line 417
        sos.commit("Finish")  # file1-3  commit third file into branch second  # line 418
        sos.changes()  # line 419
        sos.switch("test")  # file1-2, remove file3 from file tree  # line 420
        _.assertFalse(_.existsFile(3))  # removed when switching back to test  # line 421
        _.createFile("XXX")  # line 422
        out = wrapChannels(lambda: sos.status()).replace("\r", "")  # line 423
        _.assertIn("File tree has changes", out)  # line 424
        _.assertNotIn("File tree is unchanged", out)  # line 425
        _.assertIn("  * b00   'test'", out)  # line 426
        _.assertIn("    b01 'second'", out)  # line 427
        _.assertIn("(dirty)", out)  # one branch has commits  # line 428
        _.assertIn("(in sync)", out)  # the other doesn't  # line 429
        _.createFile(4, "xy")  # generate a file  # line 430
        sos.switch("second", ["--force"])  # avoids warning on uncommited changes, but keeps file4  # line 431
        _.assertFalse(_.existsFile(4))  # removed when forcedly switching back to test  # line 432
        _.assertTrue(_.existsFile(3))  # was restored from branch's revision r1  # line 433
        os.unlink("." + os.sep + "file1")  # remove old file1  # line 434
        sos.switch("test", ["--force"])  # should restore file1 and remove file3  # line 435
        _.assertTrue(_.existsFile(1))  # was restored from branch's revision r1  # line 436
        _.assertFalse(_.existsFile(3))  # was restored from branch's revision r1  # line 437

    def testAutoDetectVCS(_):  # line 439
        os.mkdir(".git")  # line 440
        sos.offline(sos.vcsBranches[sos.findSosVcsBase()[2]])  # create initial branch  # line 441
        with open(sos.metaFolder + os.sep + sos.metaFile, "r") as fd:  # line 442
            meta = fd.read()  # line 442
        _.assertTrue("\"master\"" in meta)  # line 443
        os.rmdir(".git")  # line 444

    def testUpdate(_):  # line 446
        sos.offline("trunk")  # create initial branch b0/r0  # line 447
        _.createFile(1, "x" * 100)  # line 448
        sos.commit("second")  # create b0/r1  # line 449

        sos.switch("/0")  # go back to b0/r0 - deletes file1  # line 451
        _.assertFalse(_.existsFile(1))  # line 452

        sos.update("/1")  # recreate file1  # line 454
        _.assertTrue(_.existsFile(1))  # line 455

        sos.commit("third", ["--force"])  # force because nothing to commit. should create r2 with same contents as r1, but as differential from r1, not from r0 (= no changes in meta folder)  # line 457
        _.assertTrue(os.path.exists(sos.revisionFolder(0, 2)))  # line 458
        _.assertTrue(os.path.exists(sos.revisionFolder(0, 2, file=sos.metaFile)))  # line 459
        _.assertEqual(1, len(os.listdir(sos.revisionFolder(0, 2))))  # only meta data file, no differential files  # line 460

        sos.update("/1")  # do nothing, as nothing has changed  # line 462
        _.assertTrue(_.existsFile(1))  # line 463

        _.createFile(2, "y" * 100)  # line 465
#    out = wrapChannels(() -> sos.branch("other"))  # won't comply as there are changes
#    _.assertIn("--force", out)
        sos.branch("other", options=["--force"])  # automatically including file 2 (as we are in simple mode)  # line 468
        _.assertTrue(_.existsFile(2))  # line 469
        sos.update("trunk", ["--add"])  # only add stuff  # line 470
        _.assertTrue(_.existsFile(2))  # line 471
        sos.update("trunk")  # nothing to do  # line 472
        _.assertFalse(_.existsFile(2))  # removes file not present in original branch  # line 473

        theirs = b"a\nb\nc\nd\ne\nf\ng\nh\nx\nx\nj\nk"  # line 475
        _.createFile(10, theirs)  # line 476
        mine = b"a\nc\nd\ne\ng\nf\nx\nh\ny\ny\nj"  # missing "b", inserted g, modified g->x, replace x/x -> y/y, removed k  # line 477
        _.createFile(11, mine)  # line 478
        _.assertEqual((b"a\nb\nc\nd\ne\nf\ng\nh\nx\nx\nj\nk", b"\n"), sos.merge(filename="." + os.sep + "file10", intoname="." + os.sep + "file11", mergeOperation=sos.MergeOperation.BOTH))  # completely recreated other file  # line 479
        _.assertEqual((b'a\nb\nc\nd\ne\ng\nf\ng\nh\ny\ny\nx\nx\nj\nk', b"\n"), sos.merge(filename="." + os.sep + "file10", intoname="." + os.sep + "file11", mergeOperation=sos.MergeOperation.INSERT))  # line 480

    def testUpdate2(_):  # line 482
        _.createFile("test.txt", "x" * 10)  # line 483
        sos.offline("trunk", ["--strict"])  # use strict mode, as timestamp differences are too small for testing  # line 484
        sos.branch("mod")  # line 485
        _.createFile("test.txt", "x" * 5 + "y" * 5)  # line 486
        time.sleep(FS_PRECISION)  # line 487
        sos.commit("mod")  # create b0/r1  # line 488
        sos.switch("trunk", ["--force"])  # should replace contents, force in case some other files were modified (e.g. during working on the code) TODO investigate more  # line 489
        with open("test.txt", "rb") as fd:  # line 490
            _.assertEqual(b"x" * 10, fd.read())  # line 490
        sos.update("mod")  # integrate changes TODO same with ask -> theirs  # line 491
        with open("test.txt", "rb") as fd:  # line 492
            _.assertEqual(b"x" * 5 + b"y" * 5, fd.read())  # line 492
        _.createFile("test.txt", "x" * 10)  # line 493
        mockInput(["t"], lambda: sos.update("mod", ["--ask-lines"]))  # line 494
        with open("test.txt", "rb") as fd:  # line 495
            _.assertEqual(b"x" * 5 + b"y" * 5, fd.read())  # line 495
        _.createFile("test.txt", "x" * 5 + "z" + "y" * 4)  # line 496
        sos.update("mod")  # auto-insert/removes (no intra-line conflict)  # line 497
        _.createFile("test.txt", "x" * 5 + "z" + "y" * 4)  # line 498
        mockInput(["t"], lambda: sos.update("mod", ["--ask"]))  # same as above with interaction -> use theirs (overwrite current file state)  # line 499
        with open("test.txt", "rb") as fd:  # line 500
            _.assertEqual(b"x" * 5 + b"y" * 5, fd.read())  # line 500

    def testIsTextType(_):  # line 502
        m = sos.Metadata(".")  # line 503
        m.c.texttype = ["*.x", "*.md", "*.md.*"]  # line 504
        m.c.bintype = ["*.md.confluence"]  # line 505
        _.assertTrue(m.isTextType("ab.txt"))  # line 506
        _.assertTrue(m.isTextType("./ab.txt"))  # line 507
        _.assertTrue(m.isTextType("bc/ab.txt"))  # line 508
        _.assertFalse(m.isTextType("bc/ab."))  # line 509
        _.assertTrue(m.isTextType("23_3.x.x"))  # line 510
        _.assertTrue(m.isTextType("dfg/dfglkjdf7/test.md"))  # line 511
        _.assertTrue(m.isTextType("./test.md.pdf"))  # line 512
        _.assertFalse(m.isTextType("./test_a.md.confluence"))  # line 513

    def testEolDet(_):  # line 515
        ''' Check correct end-of-line detection. '''  # line 516
        _.assertEqual(b"\n", sos.eoldet(b"a\nb"))  # line 517
        _.assertEqual(b"\r\n", sos.eoldet(b"a\r\nb\r\n"))  # line 518
        _.assertEqual(b"\r", sos.eoldet(b"\ra\rb"))  # line 519
        _.assertAllIn(["Inconsistent", "with "], wrapChannels(lambda: _.assertEqual(b"\n", sos.eoldet(b"\r\na\r\nb\n"))))  # line 520
        _.assertAllIn(["Inconsistent", "without"], wrapChannels(lambda: _.assertEqual(b"\n", sos.eoldet(b"\ra\nnb\n"))))  # line 521
        _.assertIsNone(sos.eoldet(b""))  # line 522
        _.assertIsNone(sos.eoldet(b"sdf"))  # line 523

    def testMerge(_):  # line 525
        ''' Check merge results depending on user options. '''  # line 526
        a = b"a\nb\ncc\nd"  # type: bytes  # line 527
        b = b"a\nb\nee\nd"  # type: bytes  # replaces cc by ee  # line 528
        _.assertEqual(b"a\nb\ncc\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.INSERT)[0])  # one-line block replacement using lineMerge  # line 529
        _.assertEqual(b"a\nb\neecc\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.INSERT, charMergeOperation=sos.MergeOperation.INSERT)[0])  # means insert changes from a into b, but don't replace  # line 530
        _.assertEqual(b"a\nb\n\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.INSERT, charMergeOperation=sos.MergeOperation.REMOVE)[0])  # means insert changes from a into b, but don't replace  # line 531
        _.assertEqual(b"a\nb\ncc\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.REMOVE)[0])  # one-line block replacement using lineMerge  # line 532
        _.assertEqual(b"a\nb\n\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.REMOVE, charMergeOperation=sos.MergeOperation.REMOVE)[0])  # line 533
        _.assertEqual(a, sos.merge(a, b, mergeOperation=sos.MergeOperation.BOTH)[0])  # keeps any changes in b  # line 534
        a = b"a\nb\ncc\nd"  # line 535
        b = b"a\nb\nee\nf\nd"  # replaces cc by block of two lines ee, f  # line 536
        _.assertEqual(b"a\nb\nee\nf\ncc\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.INSERT)[0])  # multi-line block replacement  # line 537
        _.assertEqual(b"a\nb\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.REMOVE)[0])  # line 538
        _.assertEqual(a, sos.merge(a, b, mergeOperation=sos.MergeOperation.BOTH)[0])  # keeps any changes in b  # line 539
# Test with change + insert
        _.assertEqual(b"a\nb fdcd d\ne", sos.merge(b"a\nb cd d\ne", b"a\nb fdd d\ne", charMergeOperation=sos.MergeOperation.INSERT)[0])  # line 541
        _.assertEqual(b"a\nb d d\ne", sos.merge(b"a\nb cd d\ne", b"a\nb fdd d\ne", charMergeOperation=sos.MergeOperation.REMOVE)[0])  # line 542
# Test interactive merge
        a = b"a\nb\nb\ne"  # block-wise replacement  # line 544
        b = b"a\nc\ne"  # line 545
        _.assertEqual(b, mockInput(["i"], lambda: sos.merge(a, b, mergeOperation=sos.MergeOperation.ASK)[0]))  # line 546
        _.assertEqual(a, mockInput(["t"], lambda: sos.merge(a, b, mergeOperation=sos.MergeOperation.ASK)[0]))  # line 547
        a = b"a\nb\ne"  # intra-line merge  # line 548
        _.assertEqual(b, mockInput(["i"], lambda: sos.merge(a, b, charMergeOperation=sos.MergeOperation.ASK)[0]))  # line 549
        _.assertEqual(a, mockInput(["t"], lambda: sos.merge(a, b, charMergeOperation=sos.MergeOperation.ASK)[0]))  # line 550

    def testMergeEol(_):  # line 552
        _.assertEqual(b"\r\n", sos.merge(b"a\nb", b"a\r\nb")[1])  # line 553
        _.assertIn("Differing EOL-styles", wrapChannels(lambda: sos.merge(b"a\nb", b"a\r\nb")))  # expects a warning  # line 554
        _.assertIn(b"a\r\nb", sos.merge(b"a\nb", b"a\r\nb")[0])  # when in doubt, use "mine" CR-LF  # line 555
        _.assertIn(b"a\nb", sos.merge(b"a\nb", b"a\r\nb", eol=True)[0])  # line 556
        _.assertEqual(b"\n", sos.merge(b"a\nb", b"a\r\nb", eol=True)[1])  # line 557

    def testPickyMode(_):  # line 559
        ''' Confirm that picky mode reset tracked patterns after commits. '''  # line 560
        sos.offline("trunk", None, ["--picky"])  # line 561
        changes = sos.changes()  # line 562
        _.assertEqual(0, len(changes.additions))  # do not list any existing file as an addition  # line 563
        sos.add(".", "./file?", ["--force"])  # line 564
        _.createFile(1, "aa")  # line 565
        sos.commit("First")  # add one file  # line 566
        _.assertEqual(2, len(os.listdir(sos.revisionFolder(0, 1))))  # line 567
        _.createFile(2, "b")  # line 568
        try:  # add nothing, because picky  # line 569
            sos.commit("Second")  # add nothing, because picky  # line 569
        except:  # line 570
            pass  # line 570
        sos.add(".", "./file?")  # line 571
        sos.commit("Third")  # line 572
        _.assertEqual(2, len(os.listdir(sos.revisionFolder(0, 2))))  # line 573
        out = wrapChannels(lambda: sos.log([])).replace("\r", "")  # line 574
        _.assertIn("  * r2", out)  # line 575
        _.createFile(3, prefix="sub")  # line 576
        sos.add("sub", "sub/file?")  # line 577
        changes = sos.changes()  # line 578
        _.assertEqual(1, len(changes.additions))  # line 579
        _.assertTrue("sub/file3" in changes.additions)  # line 580

    def testTrackedSubfolder(_):  # line 582
        ''' See if patterns for files in sub folders are picked up correctly. '''  # line 583
        os.mkdir("." + os.sep + "sub")  # line 584
        sos.offline("trunk", None, ["--track"])  # line 585
        _.createFile(1, "x")  # line 586
        _.createFile(1, "x", prefix="sub")  # line 587
        sos.add(".", "./file?")  # add glob pattern to track  # line 588
        sos.commit("First")  # line 589
        _.assertEqual(2, len(os.listdir(sos.revisionFolder(0, 1))))  # one new file + meta file  # line 590
        sos.add(".", "sub/file?")  # add glob pattern to track  # line 591
        sos.commit("Second")  # one new file + meta  # line 592
        _.assertEqual(2, len(os.listdir(sos.revisionFolder(0, 1))))  # one new file + meta file  # line 593
        os.unlink("file1")  # remove from basefolder  # line 594
        _.createFile(2, "y")  # line 595
        sos.remove(".", "sub/file?")  # line 596
        try:  # raises Exit. TODO test the "suggest a pattern" case  # line 597
            sos.remove(".", "sub/bla")  # raises Exit. TODO test the "suggest a pattern" case  # line 597
            _.fail()  # raises Exit. TODO test the "suggest a pattern" case  # line 597
        except:  # line 598
            pass  # line 598
        sos.commit("Third")  # line 599
        _.assertEqual(2, len(os.listdir(sos.revisionFolder(0, 2))))  # one new file + meta  # line 600
# TODO also check if /file1 and sub/file1 were removed from index

    def testTrackedMode(_):  # line 603
        ''' Difference in semantics vs simple mode:
          - For remote/other branch we can only know and consider tracked files, thus ignoring all complexity stemming from handling addition of untracked files.
          - For current branch, we can take into account tracked and untracked ones, in theory, but it doesn't make sense.
        In conclusion, using the union of tracking patterns from both sides to find affected files makes sense, but disallow deleting files not present in remote branch.
    '''  # line 608
        sos.offline("test", options=["--track"])  # set up repo in tracking mode (SVN- or gitless-style)  # line 609
        _.createFile(1)  # line 610
        _.createFile("a123a")  # untracked file "a123a"  # line 611
        sos.add(".", "./file?")  # add glob tracking pattern  # line 612
        sos.commit("second")  # versions "file1"  # line 613
        _.assertEqual(2, len(os.listdir(sos.revisionFolder(0, 1))))  # one new file + meta file  # line 614
        out = wrapChannels(lambda: sos.status()).replace("\r", "")  # line 615
        _.assertIn("  | ./file?", out)  # line 616

        _.createFile(2)  # untracked file "file2"  # line 618
        sos.commit("third")  # versions "file2"  # line 619
        _.assertEqual(2, len(os.listdir(sos.revisionFolder(0, 2))))  # one new file + meta file  # line 620

        os.mkdir("." + os.sep + "sub")  # line 622
        _.createFile(3, prefix="sub")  # untracked file "sub/file3"  # line 623
        sos.commit("fourth", ["--force"])  # no tracking pattern matches the subfolder  # line 624
        _.assertEqual(1, len(os.listdir(sos.revisionFolder(0, 3))))  # meta file only, no other tracked path/file  # line 625

        sos.branch("Other")  # second branch containing file1 and file2 tracked by "./file?"  # line 627
        sos.remove(".", "./file?")  # remove tracking pattern, but don't touch previously created and versioned files  # line 628
        sos.add(".", "./a*a")  # add tracking pattern  # line 629
        changes = sos.changes()  # should pick up addition only, because tracked, but not the deletion, as not tracked anymore  # line 630
        _.assertEqual(0, len(changes.modifications))  # line 631
        _.assertEqual(0, len(changes.deletions))  # not tracked anymore, but contained in version history and not removed  # line 632
        _.assertEqual(1, len(changes.additions))  # detected one addition "a123a", but won't recognize untracking files as deletion  # line 633

        sos.commit("Second_2")  # line 635
        _.assertEqual(2, len(os.listdir(sos.revisionFolder(1, 1))))  # "a123a" + meta file  # line 636
        _.existsFile(1, b"x" * 10)  # line 637
        _.existsFile(2, b"x" * 10)  # line 638

        sos.switch("test")  # go back to first branch - tracks only "file?", but not "a*a"  # line 640
        _.existsFile(1, b"x" * 10)  # line 641
        _.existsFile("a123a", b"x" * 10)  # line 642

        sos.update("Other")  # integrate tracked files and tracking pattern from second branch into working state of master branch  # line 644
        _.assertTrue(os.path.exists("." + os.sep + "file1"))  # line 645
        _.assertTrue(os.path.exists("." + os.sep + "a123a"))  # line 646

        _.createFile("axxxa")  # new file that should be tracked on "test" now that we integrated "Other"  # line 648
        sos.commit("fifth")  # create new revision after integrating updates from second branch  # line 649
        _.assertEqual(3, len(os.listdir(sos.revisionFolder(0, 4))))  # one new file from other branch + one new in current folder + meta file  # line 650
        sos.switch("Other")  # switch back to just integrated branch that tracks only "a*a" - shouldn't do anything  # line 651
        _.assertTrue(os.path.exists("." + os.sep + "file1"))  # line 652
        _.assertTrue(os.path.exists("." + os.sep + "a123a"))  # line 653
        _.assertFalse(os.path.exists("." + os.sep + "axxxa"))  # because tracked in both branches, but not present in other -> delete in file tree  # line 654
# TODO test switch --meta

    def testLsTracked(_):  # line 657
        sos.offline("test", options=["--track"])  # set up repo in tracking mode (SVN- or gitless-style)  # line 658
        _.createFile(1)  # line 659
        _.createFile("foo")  # line 660
        sos.add(".", "./file*")  # capture one file  # line 661
        sos.ls()  # line 662
        out = sos.safeSplit(wrapChannels(lambda: sos.ls()).replace("\r", ""), "\n")  # line 663
        _.assertInAny("TRK file1  (file*)", out)  # line 664
        _.assertNotInAny("... file1  (file*)", out)  # line 665
        _.assertInAny("··· foo", out)  # line 666
        out = sos.safeSplit(wrapChannels(lambda: sos.ls(options=["--patterns"])).replace("\r", ""), "\n")  # line 667
        _.assertInAny("TRK file*", out)  # line 668
        _.createFile("a", prefix="sub")  # line 669
        sos.add("sub", "sub/a")  # line 670
        sos.ls("sub")  # line 671
        _.assertIn("TRK a  (a)", sos.safeSplit(wrapChannels(lambda: sos.ls("sub")).replace("\r", ""), "\n"))  # line 672

    def testLineMerge(_):  # line 674
        _.assertEqual("xabc", sos.lineMerge("xabc", "a bd"))  # line 675
        _.assertEqual("xabxxc", sos.lineMerge("xabxxc", "a bd"))  # line 676
        _.assertEqual("xa bdc", sos.lineMerge("xabc", "a bd", mergeOperation=sos.MergeOperation.INSERT))  # line 677
        _.assertEqual("ab", sos.lineMerge("xabc", "a bd", mergeOperation=sos.MergeOperation.REMOVE))  # line 678

    def testCompression(_):  # TODO test output ratio/advantage, also depending on compress flag set or not  # line 680
        _.createFile(1)  # line 681
        sos.offline("master", options=["--force"])  # line 682
        out = wrapChannels(lambda: sos.changes(options=['--progress'])).replace("\r", "").split("\n")  # line 683
        _.assertFalse(any(("Compression advantage" in line for line in out)))  # simple mode should always print this to stdout  # line 684
        _.assertTrue(_.existsFile(sos.revisionFolder(0, 0, file="b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa"), b"x" * 10))  # line 685
        setRepoFlag("compress", True)  # was plain = uncompressed before  # line 686
        _.createFile(2)  # line 687
        out = wrapChannels(lambda: sos.commit("Added file2", options=['--progress'])).replace("\r", "").split("\n")  # line 688
        _.assertTrue(any(("Compression advantage" in line for line in out)))  # line 689
        _.assertTrue(_.existsFile(sos.revisionFolder(0, 1, file="03b69bc801ae11f1ff2a71a50f165996d0ad681b4f822df13329a27e53f0fcd2")))  # exists  # line 690
        _.assertFalse(_.existsFile(sos.revisionFolder(0, 1, file="03b69bc801ae11f1ff2a71a50f165996d0ad681b4f822df13329a27e53f0fcd2"), b"x" * 10))  # but is compressed instead  # line 691

    def testLocalConfig(_):  # line 693
        sos.offline("bla", options=[])  # line 694
        try:  # line 695
            sos.config(["set", "ignores", "one;two"], options=["--local"])  # line 695
        except SystemExit as E:  # line 696
            _.assertEqual(0, E.code)  # line 696
        _.assertTrue(checkRepoFlag("ignores", value=["one", "two"]))  # line 697

    def testConfigVariations(_):  # line 699
        def makeRepo():  # line 700
            try:  # line 701
                os.unlink("file1")  # line 701
            except:  # line 702
                pass  # line 702
            sos.offline("master", options=["--force"])  # line 703
            _.createFile(1)  # line 704
            sos.commit("Added file1")  # line 705
        try:  # line 706
            sos.config(["set", "strict", "on"])  # line 706
        except SystemExit as E:  # line 707
            _.assertEqual(0, E.code)  # line 707
        makeRepo()  # line 708
        _.assertTrue(checkRepoFlag("strict", True))  # line 709
        try:  # line 710
            sos.config(["set", "strict", "off"])  # line 710
        except SystemExit as E:  # line 711
            _.assertEqual(0, E.code)  # line 711
        makeRepo()  # line 712
        _.assertTrue(checkRepoFlag("strict", False))  # line 713
        try:  # line 714
            sos.config(["set", "strict", "yes"])  # line 714
        except SystemExit as E:  # line 715
            _.assertEqual(0, E.code)  # line 715
        makeRepo()  # line 716
        _.assertTrue(checkRepoFlag("strict", True))  # line 717
        try:  # line 718
            sos.config(["set", "strict", "no"])  # line 718
        except SystemExit as E:  # line 719
            _.assertEqual(0, E.code)  # line 719
        makeRepo()  # line 720
        _.assertTrue(checkRepoFlag("strict", False))  # line 721
        try:  # line 722
            sos.config(["set", "strict", "1"])  # line 722
        except SystemExit as E:  # line 723
            _.assertEqual(0, E.code)  # line 723
        makeRepo()  # line 724
        _.assertTrue(checkRepoFlag("strict", True))  # line 725
        try:  # line 726
            sos.config(["set", "strict", "0"])  # line 726
        except SystemExit as E:  # line 727
            _.assertEqual(0, E.code)  # line 727
        makeRepo()  # line 728
        _.assertTrue(checkRepoFlag("strict", False))  # line 729
        try:  # line 730
            sos.config(["set", "strict", "true"])  # line 730
        except SystemExit as E:  # line 731
            _.assertEqual(0, E.code)  # line 731
        makeRepo()  # line 732
        _.assertTrue(checkRepoFlag("strict", True))  # line 733
        try:  # line 734
            sos.config(["set", "strict", "false"])  # line 734
        except SystemExit as E:  # line 735
            _.assertEqual(0, E.code)  # line 735
        makeRepo()  # line 736
        _.assertTrue(checkRepoFlag("strict", False))  # line 737
        try:  # line 738
            sos.config(["set", "strict", "enable"])  # line 738
        except SystemExit as E:  # line 739
            _.assertEqual(0, E.code)  # line 739
        makeRepo()  # line 740
        _.assertTrue(checkRepoFlag("strict", True))  # line 741
        try:  # line 742
            sos.config(["set", "strict", "disable"])  # line 742
        except SystemExit as E:  # line 743
            _.assertEqual(0, E.code)  # line 743
        makeRepo()  # line 744
        _.assertTrue(checkRepoFlag("strict", False))  # line 745
        try:  # line 746
            sos.config(["set", "strict", "enabled"])  # line 746
        except SystemExit as E:  # line 747
            _.assertEqual(0, E.code)  # line 747
        makeRepo()  # line 748
        _.assertTrue(checkRepoFlag("strict", True))  # line 749
        try:  # line 750
            sos.config(["set", "strict", "disabled"])  # line 750
        except SystemExit as E:  # line 751
            _.assertEqual(0, E.code)  # line 751
        makeRepo()  # line 752
        _.assertTrue(checkRepoFlag("strict", False))  # line 753
        try:  # line 754
            sos.config(["set", "strict", "nope"])  # line 754
            _.fail()  # line 754
        except SystemExit as E:  # line 755
            _.assertEqual(1, E.code)  # line 755

    def testLsSimple(_):  # line 757
        _.createFile(1)  # line 758
        _.createFile("foo")  # line 759
        _.createFile("ign1")  # line 760
        _.createFile("ign2")  # line 761
        sos.offline("test")  # set up repo in tracking mode (SVN- or gitless-style)  # line 762
        try:  # define an ignore pattern  # line 763
            sos.config(["set", "ignores", "ign1"])  # define an ignore pattern  # line 763
        except SystemExit as E:  # line 764
            _.assertEqual(0, E.code)  # line 764
        try:  # additional ignore pattern  # line 765
            sos.config(["add", "ignores", "ign2"])  # additional ignore pattern  # line 765
        except SystemExit as E:  # line 766
            _.assertEqual(0, E.code)  # line 766
        try:  # define a list of ignore patterns  # line 767
            sos.config(["set", "ignoresWhitelist", "ign1;ign2"])  # define a list of ignore patterns  # line 767
        except SystemExit as E:  # line 768
            _.assertEqual(0, E.code)  # line 768
        out = wrapChannels(lambda: sos.config(["show"])).replace("\r", "")  # line 769
        _.assertIn("             ignores [global]  ['ign1', 'ign2']", out)  # line 770
        out = sos.safeSplit(wrapChannels(lambda: sos.ls()).replace("\r", ""), "\n")  # line 771
        _.assertInAny('··· file1', out)  # line 772
        _.assertInAny('··· ign1', out)  # line 773
        _.assertInAny('··· ign2', out)  # line 774
        try:  # line 775
            sos.config(["rm", "foo", "bar"])  # line 775
            _.fail()  # line 775
        except SystemExit as E:  # line 776
            _.assertEqual(1, E.code)  # line 776
        try:  # line 777
            sos.config(["rm", "ignores", "foo"])  # line 777
            _.fail()  # line 777
        except SystemExit as E:  # line 778
            _.assertEqual(1, E.code)  # line 778
        try:  # line 779
            sos.config(["rm", "ignores", "ign1"])  # line 779
        except SystemExit as E:  # line 780
            _.assertEqual(0, E.code)  # line 780
        try:  # remove ignore pattern  # line 781
            sos.config(["unset", "ignoresWhitelist"])  # remove ignore pattern  # line 781
        except SystemExit as E:  # line 782
            _.assertEqual(0, E.code)  # line 782
        out = sos.safeSplit(wrapChannels(lambda: sos.ls()).replace("\r", ""), "\n")  # line 783
        _.assertInAny('··· ign1', out)  # line 784
        _.assertInAny('IGN ign2', out)  # line 785
        _.assertNotInAny('··· ign2', out)  # line 786

    def testWhitelist(_):  # line 788
# TODO test same for simple mode
        _.createFile(1)  # line 790
        sos.defaults.ignores[:] = ["file*"]  # replace in-place  # line 791
        sos.offline("xx", options=["--track", "--strict"])  # because nothing to commit due to ignore pattern  # line 792
        sos.add(".", "./file*")  # add tracking pattern for "file1"  # line 793
        sos.commit(options=["--force"])  # attempt to commit the file  # line 794
        _.assertEqual(1, len(os.listdir(sos.revisionFolder(0, 1))))  # only meta data, file1 was ignored  # line 795
        try:  # Exit because dirty  # line 796
            sos.online()  # Exit because dirty  # line 796
            _.fail()  # Exit because dirty  # line 796
        except:  # exception expected  # line 797
            pass  # exception expected  # line 797
        _.createFile("x2")  # add another change  # line 798
        sos.add(".", "./x?")  # add tracking pattern for "file1"  # line 799
        try:  # force beyond dirty flag check  # line 800
            sos.online(["--force"])  # force beyond dirty flag check  # line 800
            _.fail()  # force beyond dirty flag check  # line 800
        except:  # line 801
            pass  # line 801
        sos.online(["--force", "--force"])  # force beyond file tree modifications check  # line 802
        _.assertFalse(os.path.exists(sos.metaFolder))  # line 803

        _.createFile(1)  # line 805
        sos.defaults.ignoresWhitelist[:] = ["file*"]  # line 806
        sos.offline("xx", None, ["--track"])  # line 807
        sos.add(".", "./file*")  # line 808
        sos.commit()  # should NOT ask for force here  # line 809
        _.assertEqual(2, len(os.listdir(sos.revisionFolder(0, 1))))  # meta data and "file1", file1 was whitelisted  # line 810

    def testRemove(_):  # line 812
        _.createFile(1, "x" * 100)  # line 813
        sos.offline("trunk")  # line 814
        try:  # line 815
            sos.destroy("trunk")  # line 815
            _fail()  # line 815
        except:  # line 816
            pass  # line 816
        _.createFile(2, "y" * 10)  # line 817
        sos.branch("added")  # creates new branch, writes repo metadata, and therefore creates backup copy  # line 818
        sos.destroy("trunk")  # line 819
        _.assertAllIn([sos.metaFile, sos.metaBack, "b0_last", "b1"], os.listdir("." + os.sep + sos.metaFolder))  # line 820
        _.assertTrue(os.path.exists("." + os.sep + sos.metaFolder + os.sep + "b1"))  # line 821
        _.assertFalse(os.path.exists("." + os.sep + sos.metaFolder + os.sep + "b0"))  # line 822
        sos.branch("next")  # line 823
        _.createFile(3, "y" * 10)  # make a change  # line 824
        sos.destroy("added", "--force")  # should succeed  # line 825

    def testUsage(_):  # line 827
        try:  # TODO expect sys.exit(0)  # line 828
            sos.usage()  # TODO expect sys.exit(0)  # line 828
            _.fail()  # TODO expect sys.exit(0)  # line 828
        except:  # line 829
            pass  # line 829
        try:  # line 830
            sos.usage(short=True)  # line 830
            _.fail()  # line 830
        except:  # line 831
            pass  # line 831

    def testOnlyExcept(_):  # line 833
        ''' Test blacklist glob rules. '''  # line 834
        sos.offline(options=["--track"])  # line 835
        _.createFile("a.1")  # line 836
        _.createFile("a.2")  # line 837
        _.createFile("b.1")  # line 838
        _.createFile("b.2")  # line 839
        sos.add(".", "./a.?")  # line 840
        sos.add(".", "./?.1", negative=True)  # line 841
        out = wrapChannels(lambda: sos.commit())  # line 842
        _.assertIn("ADD ./a.2", out)  # line 843
        _.assertNotIn("ADD ./a.1", out)  # line 844
        _.assertNotIn("ADD ./b.1", out)  # line 845
        _.assertNotIn("ADD ./b.2", out)  # line 846

    def testOnly(_):  # line 848
        _.assertEqual((_coconut.frozenset(("./A", "x/B")), _coconut.frozenset(("./C",))), sos.parseOnlyOptions(".", ["abc", "def", "--only", "A", "--x", "--only", "x/B", "--except", "C", "--only"]))  # line 849
        _.assertEqual(_coconut.frozenset(("B",)), sos.conditionalIntersection(_coconut.frozenset(("A", "B", "C")), _coconut.frozenset(("B", "D"))))  # line 850
        _.assertEqual(_coconut.frozenset(("B", "D")), sos.conditionalIntersection(_coconut.frozenset(), _coconut.frozenset(("B", "D"))))  # line 851
        _.assertEqual(_coconut.frozenset(("B", "D")), sos.conditionalIntersection(None, _coconut.frozenset(("B", "D"))))  # line 852
        sos.offline(options=["--track", "--strict"])  # line 853
        _.createFile(1)  # line 854
        _.createFile(2)  # line 855
        sos.add(".", "./file1")  # line 856
        sos.add(".", "./file2")  # line 857
        sos.commit(onlys=_coconut.frozenset(("./file1",)))  # line 858
        _.assertEqual(2, len(os.listdir(sos.revisionFolder(0, 1))))  # only meta and file1  # line 859
        sos.commit()  # adds also file2  # line 860
        _.assertEqual(2, len(os.listdir(sos.revisionFolder(0, 2))))  # only meta and file1  # line 861
        _.createFile(1, "cc")  # modify both files  # line 862
        _.createFile(2, "dd")  # line 863
        try:  # line 864
            sos.config(["set", "texttype", "file2"])  # line 864
        except SystemExit as E:  # line 865
            _.assertEqual(0, E.code)  # line 865
        changes = sos.changes(excps=_coconut.frozenset(("./file1",)))  # line 866
        _.assertEqual(1, len(changes.modifications))  # only file2  # line 867
        _.assertTrue("./file2" in changes.modifications)  # line 868
        _.assertAllIn(["DIF ./file2", "<No newline>"], wrapChannels(lambda: sos.diff(onlys=_coconut.frozenset(("./file2",)))))  # line 869
        _.assertAllNotIn(["MOD ./file1", "DIF ./file1", "MOD ./file2"], wrapChannels(lambda: sos.diff(onlys=_coconut.frozenset(("./file2",)))))  # line 870

    def testDiff(_):  # line 872
        try:  # manually mark this file as "textual"  # line 873
            sos.config(["set", "texttype", "file1"])  # manually mark this file as "textual"  # line 873
        except SystemExit as E:  # line 874
            _.assertEqual(0, E.code)  # line 874
        sos.offline(options=["--strict"])  # line 875
        _.createFile(1)  # line 876
        _.createFile(2)  # line 877
        sos.commit()  # line 878
        _.createFile(1, "sdfsdgfsdf")  # line 879
        _.createFile(2, "12343")  # line 880
        sos.commit()  # line 881
        _.createFile(1, "foobar")  # line 882
        _.createFile(3)  # line 883
        out = wrapChannels(lambda: sos.diff("/-2"))  # compare with r1 (second counting from last which is r2)  # line 884
        _.assertIn("ADD ./file3", out)  # line 885
        _.assertAllIn(["MOD ./file2", "DIF ./file1  <No newline>", "- | 0001 |xxxxxxxxxx|", "+ | 0000 |foobar|"], out)  # line 886
        _.assertAllNotIn(["MOD ./file1", "DIF ./file1"], wrapChannels(lambda: sos.diff("/-2", onlys=_coconut.frozenset(("./file2",)))))  # line 887

    def testReorderRenameActions(_):  # line 889
        result = sos.reorderRenameActions([("123", "312"), ("312", "132"), ("321", "123")], exitOnConflict=False)  # type: Tuple[str, str]  # line 890
        _.assertEqual([("312", "132"), ("123", "312"), ("321", "123")], result)  # line 891
        try:  # line 892
            sos.reorderRenameActions([("123", "312"), ("312", "123")], exitOnConflict=True)  # line 892
            _.fail()  # line 892
        except:  # line 893
            pass  # line 893

    def testMove(_):  # line 895
        sos.offline(options=["--strict", "--track"])  # line 896
        _.createFile(1)  # line 897
        sos.add(".", "./file?")  # line 898
# test source folder missing
        try:  # line 900
            sos.move("sub", "sub/file?", ".", "?file")  # line 900
            _.fail()  # line 900
        except:  # line 901
            pass  # line 901
# test target folder missing: create it
        sos.move(".", "./file?", "sub", "sub/file?")  # line 903
        _.assertTrue(os.path.exists("sub"))  # line 904
        _.assertTrue(os.path.exists("sub/file1"))  # line 905
        _.assertFalse(os.path.exists("file1"))  # line 906
# test move
        sos.move("sub", "sub/file?", ".", "./?file")  # line 908
        _.assertTrue(os.path.exists("1file"))  # line 909
        _.assertFalse(os.path.exists("sub/file1"))  # line 910
# test nothing matches source pattern
        try:  # line 912
            sos.move(".", "a*", ".", "b*")  # line 912
            _.fail()  # line 912
        except:  # line 913
            pass  # line 913
        sos.add(".", "*")  # anything pattern  # line 914
        try:  # TODO check that alternative pattern "*" was suggested (1 hit)  # line 915
            sos.move(".", "a*", ".", "b*")  # TODO check that alternative pattern "*" was suggested (1 hit)  # line 915
            _.fail()  # TODO check that alternative pattern "*" was suggested (1 hit)  # line 915
        except:  # line 916
            pass  # line 916
# test rename no conflict
        _.createFile(1)  # line 918
        _.createFile(2)  # line 919
        _.createFile(3)  # line 920
        sos.add(".", "./file*")  # line 921
        try:  # define an ignore pattern  # line 922
            sos.config(["set", "ignores", "file3;file4"])  # define an ignore pattern  # line 922
        except SystemExit as E:  # line 923
            _.assertEqual(0, E.code)  # line 923
        try:  # line 924
            sos.config(["set", "ignoresWhitelist", "file3"])  # line 924
        except SystemExit as E:  # line 925
            _.assertEqual(0, E.code)  # line 925
        sos.move(".", "./file*", ".", "fi*le")  # line 926
        _.assertTrue(all((os.path.exists("fi%dle" % i) for i in range(1, 4))))  # line 927
        _.assertFalse(os.path.exists("fi4le"))  # line 928
# test rename solvable conflicts
        [_.createFile("%s-%s-%s" % tuple((c for c in n))) for n in ["312", "321", "123", "231"]]  # line 930
#    sos.move("?-?-?")
# test rename unsolvable conflicts
# test --soft option
        sos.remove(".", "./?file")  # was renamed before  # line 934
        sos.add(".", "./?a?b", ["--force"])  # line 935
        sos.move(".", "./?a?b", ".", "./a?b?", ["--force", "--soft"])  # line 936
        _.createFile("1a2b")  # should not be tracked  # line 937
        _.createFile("a1b2")  # should be tracked  # line 938
        sos.commit()  # line 939
        _.assertEqual(2, len(os.listdir(sos.revisionFolder(0, 1))))  # line 940
        _.assertTrue(os.path.exists(sos.revisionFolder(0, 1, file="93b38f90892eb5c57779ca9c0b6fbdf6774daeee3342f56f3e78eb2fe5336c50")))  # a1b2  # line 941
        _.createFile("1a1b1")  # line 942
        _.createFile("1a1b2")  # line 943
        sos.add(".", "?a?b*")  # line 944
        _.assertIn("not unique", wrapChannels(lambda: sos.move(".", "?a?b*", ".", "z?z?")))  # should raise error due to same target name  # line 945
# TODO only rename if actually any files are versioned? or simply what is alife?
# TODO add test if two single question marks will be moved into adjacent characters

    def testHashCollision(_):  # line 949
        sos.offline()  # line 950
        _.createFile(1)  # line 951
        os.mkdir(sos.revisionFolder(0, 1))  # line 952
        _.createFile("b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa", prefix=sos.revisionFolder(0, 1))  # line 953
        _.createFile(1)  # line 954
        try:  # should exit with error due to collision detection  # line 955
            sos.commit()  # should exit with error due to collision detection  # line 955
            _.fail()  # should exit with error due to collision detection  # line 955
        except SystemExit as E:  # TODO will capture exit(0) which is wrong, change to check code in all places  # line 956
            _.assertEqual(1, E.code)  # TODO will capture exit(0) which is wrong, change to check code in all places  # line 956

    def testFindBase(_):  # line 958
        old = os.getcwd()  # line 959
        try:  # line 960
            os.mkdir("." + os.sep + ".git")  # line 961
            os.makedirs("." + os.sep + "a" + os.sep + sos.metaFolder)  # line 962
            os.makedirs("." + os.sep + "a" + os.sep + "b")  # line 963
            os.chdir("a" + os.sep + "b")  # line 964
            s, vcs, cmd = sos.findSosVcsBase()  # line 965
            _.assertIsNotNone(s)  # line 966
            _.assertIsNotNone(vcs)  # line 967
            _.assertEqual("git", cmd)  # line 968
        finally:  # line 969
            os.chdir(old)  # line 969

# TODO test command line operation --sos vs. --vcs
# check exact output instead of only expected exception/fail

# TODO test +++ --- in diff
# TODO test +01/-02/*..
# TODO tests for loadcommit redirection
# TODO test wrong branch/revision after fast branching, would raise exception for -1 otherwise

if __name__ == '__main__':  # line 979
    logging.basicConfig(level=logging.DEBUG if '-v' in sys.argv or "true" in [os.getenv("DEBUG", "false").strip().lower(), os.getenv("CI", "false").strip().lower()] else logging.INFO, stream=sys.stderr, format="%(asctime)-23s %(levelname)-8s %(name)s:%(lineno)d | %(message)s" if '-v' in sys.argv or os.getenv("DEBUG", "false") == "true" else "%(message)s")  # line 980
    c = configr.Configr("sos")  # line 981
    c.loadSettings()  # line 982
    if len(c.keys()) > 0:  # line 983
        sos.Exit("Cannot run test suite with existing local SOS user configuration (would affect results)")  # line 983
    unittest.main(testRunner=debugTestRunner() if '-v' in sys.argv and not os.getenv("CI", "false").lower() == "true" else None)  # warnings = "ignore")  # line 984



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0x27964731

# Compiled with Coconut version 1.3.1-post_dev18 [Dead Parrot]

# Coconut Header: -------------------------------------------------------------

import sys as _coconut_sys, os.path as _coconut_os_path
_coconut_file_path = _coconut_os_path.dirname(_coconut_os_path.abspath(__file__))
_coconut_sys.path.insert(0, _coconut_file_path)
from __coconut__ import _coconut, _coconut_NamedTuple, _coconut_MatchError, _coconut_tail_call, _coconut_tco, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_pipe, _coconut_star_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial
from __coconut__ import *
_coconut_sys.path.remove(_coconut_file_path)

# Compiled Coconut: -----------------------------------------------------------

# Copyright Arne Bachmann
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

# Utiliy functions
import bz2  # line 5
import codecs  # line 5
import difflib  # line 5
import hashlib  # line 5
import logging  # line 5
import os  # line 5
import re  # line 5
sys = _coconut_sys  # line 5
import time  # line 5
START_TIME = time.time()  # line 5
try:  # line 6
    import enum  # line 6
except:  # line 7
    raise Exception("Please install SOS via 'pip install -U sos-vcs[backport]' to get enum support for Python versions prior 3.4")  # line 7
from typing import Generic  # line 8
from typing import TypeVar  # line 8
Number = TypeVar("Number", int, float)  # line 9
try:  # line 10
    import wcwidth  # line 10
except:  # optional dependency  # line 11
    pass  # optional dependency  # line 11


verbose = os.environ.get("DEBUG", "False").lower() == "true" or '--verbose' in sys.argv or '-v' in sys.argv  # imported from utility, and only modified here  # line 14

# Classes
class Accessor(dict):  # line 17
    ''' Dictionary with attribute access. Writing only supported via dictionary access. '''  # line 18
    def __init__(_, mapping: 'Dict[str, Any]'):  # line 19
        dict.__init__(_, mapping)  # line 19
    @_coconut_tco  # line 20
    def __getattribute__(_, name: 'str') -> 'Any':  # line 20
        try:  # line 21
            return _[name]  # line 21
        except:  # line 22
            return _coconut_tail_call(dict.__getattribute__, _, name)  # line 22

class Counter(Generic[Number]):  # line 24
    ''' A simple counter. Can be augmented to return the last value instead. '''  # line 25
    def __init__(_, initial: 'Number'=0):  # line 26
        _.value = initial  # type: Number  # line 26
    def inc(_, by: 'Number'=1) -> 'Number':  # line 27
        _.value += by  # line 27
        return _.value  # line 27

class ProgressIndicator(Counter):  # line 29
    ''' Manages a rotating progress indicator. '''  # line 30
    def __init__(_, callback: 'Optional[_coconut.typing.Callable[[str], None]]'=None):  # line 31
        super(ProgressIndicator, _).__init__(-1)  # line 31
        _.timer = time.time()  # type: float  # line 31
        _.callback = callback  # type: Optional[_coconut.typing.Callable[[str], None]]  # line 31
    def getIndicator(_) -> '_coconut.typing.Optional[str]':  # line 32
        ''' Returns a value only if a certain time has passed. '''  # line 33
        newtime = time.time()  # line 34
        if newtime - _.timer < .1:  # line 35
            return None  # line 35
        _.timer = newtime  # line 36
        sign = PROGRESS_MARKER[int(_.inc() % 4)]  # type: str  # line 37
        if _.callback:  # line 38
            _.callback(sign)  # line 38
        return sign  # line 39

class Logger:  # line 41
    ''' Logger that supports many items. '''  # line 42
    def __init__(_, log):  # line 43
        _._log = log  # line 43
    def debug(_, *s):  # line 44
        _._log.debug(sjoin(*s))  # line 44
    def info(_, *s):  # line 45
        _._log.info(sjoin(*s))  # line 45
    def warn(_, *s):  # line 46
        _._log.warning(sjoin(*s))  # line 46
    def error(_, *s):  # line 47
        _._log.error(sjoin(*s))  # line 47


# Constants
_log = Logger(logging.getLogger(__name__))  # line 51
debug, info, warn, error = _log.debug, _log.info, _log.warn, _log.error  # line 51
CONFIGURABLE_FLAGS = ["strict", "track", "picky", "compress", "useChangesCommand"]  # type: List[str]  # line 52
CONFIGURABLE_LISTS = ["texttype", "bintype", "ignores", "ignoreDirs", "ignoresWhitelist", "ignoreDirsWhitelist"]  # type: List[str]  # line 53
GLOBAL_LISTS = ["ignores", "ignoreDirs", "ignoresWhitelist", "ignoreDirsWhitelist"]  # type: List[str]  # line 54
TRUTH_VALUES = ["true", "yes", "on", "1", "enable", "enabled"]  # type: List[str]  # all lower-case normalized  # line 55
FALSE_VALUES = ["false", "no", "off", "0", "disable", "disabled"]  # type: List[str]  # line 56
PROGRESS_MARKER = ["|", "/", "-", "\\"]  # type: List[str]  # line 57
BACKUP_SUFFIX = "_last"  # type: str  # line 58
metaFolder = ".sos"  # type: str  # line 59
DUMP_FILE = metaFolder + ".zip"  # type: str  # line 60
metaFile = ".meta"  # type: str  # line 61
metaBack = metaFile + BACKUP_SUFFIX  # type: str  # line 62
MEBI = 1 << 20  # type: int  # line 63
bufSize = MEBI  # type: int  # line 64
UTF8 = "utf_8"  # type: str  # early used constant, not defined in standard library  # line 65
SVN = "svn"  # type: str  # line 66
SLASH = "/"  # type: str  # line 67
METADATA_FORMAT = 1  # type: int  # counter for incompatible consecutive formats  # line 68
vcsFolders = {".svn": SVN, ".git": "git", ".bzr": "bzr", ".hg": "hg", ".fslckout": "fossil", "_FOSSIL_": "fossil", ".CVS": "cvs"}  # type: Dict[str, str]  # line 69
vcsBranches = {SVN: "trunk", "git": "master", "bzr": "trunk", "hg": "default", "fossil": None, "cvs": None}  # type: Dict[str, _coconut.typing.Optional[str]]  # line 70
NL_NAMES = {None: "<No newline>", b"\r\n": "<CR+LF>", b"\n\r": "<LF+CR>", b"\n": "<LF>", b"\r": "<CR>"}  # type: Dict[bytes, str]  # line 71
defaults = Accessor({"strict": False, "track": False, "picky": False, "compress": False, "useChangesCommand": False, "texttype": ["*.md", "*.coco", "*.py", "*.pyi", "*.pth"], "bintype": [], "ignoreDirs": [".*", "__pycache__", ".mypy_cache"], "ignoreDirsWhitelist": [], "ignores": ["__coconut__.py", "*.bak", "*.py[cdo]", "*.class", ".fslckout", "_FOSSIL_", "*%s" % DUMP_FILE], "ignoresWhitelist": []})  # type: Accessor  # line 72


# Enums
MergeOperation = enum.Enum("MergeOperation", {"INSERT": 1, "REMOVE": 2, "BOTH": 3, "ASK": 4})  # insert remote changes into current, remove remote deletions from current, do both (replicates remote state), or ask per block  # line 84
MergeBlockType = enum.Enum("MergeBlockType", "KEEP INSERT REMOVE REPLACE MOVE")  # modify = intra-line changes, replace = full block replacement  # line 85


# Value types
class BranchInfo(_coconut_NamedTuple("BranchInfo", [("number", 'int'), ("ctime", 'int'), ("name", '_coconut.typing.Optional[str]'), ("inSync", 'bool'), ("tracked", 'List[str]'), ("untracked", 'List[str]'), ("parent", '_coconut.typing.Optional[int]'), ("revision", '_coconut.typing.Optional[int]')])):  # tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place  # line 89
    __slots__ = ()  # tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place  # line 89
    __ne__ = _coconut.object.__ne__  # tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place  # line 89
    def __new__(_cls, number, ctime, name=None, inSync=False, tracked=[], untracked=[], parent=None, revision=None):  # tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place  # line 89
        return _coconut.tuple.__new__(_cls, (number, ctime, name, inSync, tracked, untracked, parent, revision))  # tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place  # line 89
# tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place
class CommitInfo(_coconut_NamedTuple("CommitInfo", [("number", 'int'), ("ctime", 'int'), ("message", '_coconut.typing.Optional[str]')])):  # line 90
    __slots__ = ()  # line 90
    __ne__ = _coconut.object.__ne__  # line 90
    def __new__(_cls, number, ctime, message=None):  # line 90
        return _coconut.tuple.__new__(_cls, (number, ctime, message))  # line 90

class PathInfo(_coconut_NamedTuple("PathInfo", [("nameHash", 'str'), ("size", '_coconut.typing.Optional[int]'), ("mtime", 'int'), ("hash", '_coconut.typing.Optional[str]')])):  # size == None means deleted in this revision  # line 91
    __slots__ = ()  # size == None means deleted in this revision  # line 91
    __ne__ = _coconut.object.__ne__  # size == None means deleted in this revision  # line 91
class ChangeSet(_coconut_NamedTuple("ChangeSet", [("additions", 'Dict[str, PathInfo]'), ("deletions", 'Dict[str, PathInfo]'), ("modifications", 'Dict[str, PathInfo]'), ("moves", 'Dict[str, Tuple[str, PathInfo]]')])):  # avoid default assignment of {} as it leads to runtime errors (contains data on init for unknown reason)  # line 92
    __slots__ = ()  # avoid default assignment of {} as it leads to runtime errors (contains data on init for unknown reason)  # line 92
    __ne__ = _coconut.object.__ne__  # avoid default assignment of {} as it leads to runtime errors (contains data on init for unknown reason)  # line 92
class Range(_coconut_NamedTuple("Range", [("tipe", 'MergeBlockType'), ("indexes", '_coconut.typing.Sequence[int]')])):  # MergeBlockType[1,2,4], line number, length  # TODO use enum  # line 93
    __slots__ = ()  # MergeBlockType[1,2,4], line number, length  # TODO use enum  # line 93
    __ne__ = _coconut.object.__ne__  # MergeBlockType[1,2,4], line number, length  # TODO use enum  # line 93
class MergeBlock(_coconut_NamedTuple("MergeBlock", [("tipe", 'MergeBlockType'), ("lines", 'List[str]'), ("line", 'int'), ("replaces", '_coconut.typing.Optional[MergeBlock]'), ("changes", '_coconut.typing.Optional[Range]')])):  # line 94
    __slots__ = ()  # line 94
    __ne__ = _coconut.object.__ne__  # line 94
    def __new__(_cls, tipe, lines, line, replaces=None, changes=None):  # line 94
        return _coconut.tuple.__new__(_cls, (tipe, lines, line, replaces, changes))  # line 94

class GlobBlock(_coconut_NamedTuple("GlobBlock", [("isLiteral", 'bool'), ("content", 'str'), ("index", 'int')])):  # for file pattern rename/move matching  # line 95
    __slots__ = ()  # for file pattern rename/move matching  # line 95
    __ne__ = _coconut.object.__ne__  # for file pattern rename/move matching  # line 95
class GlobBlock2(_coconut_NamedTuple("GlobBlock2", [("isLiteral", 'bool'), ("content", 'str'), ("matches", 'str')])):  # matching file pattern and input filename for translation  # line 96
    __slots__ = ()  # matching file pattern and input filename for translation  # line 96
    __ne__ = _coconut.object.__ne__  # matching file pattern and input filename for translation  # line 96
DataType = TypeVar("DataType", BranchInfo, ChangeSet, MergeBlock, PathInfo)  # line 97


# Functions
def printo(s: 'str'="", nl: 'str'="\n"):  # PEP528 compatibility  # line 101
    sys.stdout.buffer.write((s + nl).encode(sys.stdout.encoding, 'backslashreplace'))  # PEP528 compatibility  # line 101
    sys.stdout.flush()  # PEP528 compatibility  # line 101
def printe(s: 'str'="", nl: 'str'="\n"):  # line 102
    sys.stderr.buffer.write((s + nl).encode(sys.stderr.encoding, 'backslashreplace'))  # line 102
    sys.stderr.flush()  # line 102
@_coconut_tco  # for py->os access of writing filenames  # PEP 529 compatibility  # line 103
def encode(s: 'str') -> 'bytes':  # for py->os access of writing filenames  # PEP 529 compatibility  # line 103
    return _coconut_tail_call(os.fsencode, s)  # for py->os access of writing filenames  # PEP 529 compatibility  # line 103
@_coconut_tco  # for os->py access of reading filenames  # line 104
def decode(b: 'bytes') -> 'str':  # for os->py access of reading filenames  # line 104
    return _coconut_tail_call(os.fsdecode, b)  # for os->py access of reading filenames  # line 104
try:  # line 105
    import chardet  # https://github.com/chardet/chardet  # line 106
    def detectEncoding(binary: 'bytes') -> 'str':  # line 107
        return chardet.detect(binary)["encoding"]  # line 107
except:  # line 108
    def detectEncoding(binary: 'bytes') -> 'str':  # Guess the encoding  # line 109
        ''' Fallback if chardet library missing. '''  # line 110
        try:  # line 111
            binary.decode(UTF8)  # line 111
            return UTF8  # line 111
        except UnicodeError:  # line 112
            pass  # line 112
        try:  # line 113
            binary.decode("utf_16")  # line 113
            return "utf_16"  # line 113
        except UnicodeError:  # line 114
            pass  # line 114
        try:  # line 115
            binary.decode("cp1252")  # line 115
            return "cp1252"  # line 115
        except UnicodeError:  # line 116
            pass  # line 116
        return "ascii"  # this code will never be reached, as above is an 8-bit charset that always matches  # line 117

def tryOrDefault(func: '_coconut.typing.Callable[[], Any]', default: 'Any') -> 'Any':  # line 119
    try:  # line 120
        return func()  # line 120
    except:  # line 121
        return default  # line 121

def tryOrIgnore(func: '_coconut.typing.Callable[[], Any]') -> 'None':  # handle with care!  # line 123
    try:  # line 124
        func()  # line 124
    except:  # line 125
        pass  # line 125

@_coconut_tco  # line 127
def wcswidth(string: 'str') -> 'int':  # line 127
    l = 0  # type: int  # line 128
    try:  # line 129
        l = wcwidth.wcswitdh(string)  # line 130
        return len(string) if l < 0 else l  # line 131
    finally:  # line 132
        return _coconut_tail_call(len, string)  # line 132

def removePath(key: 'str', value: 'str') -> 'str':  # line 134
    ''' Cleanup of user-specified global file patterns. '''  # line 135
    return value if value in GLOBAL_LISTS or SLASH not in value else value[value.rindex(SLASH) + 1:]  # line 136

def conditionalIntersection(a: '_coconut.typing.Optional[FrozenSet[str]]', b: 'FrozenSet[str]') -> 'FrozenSet[str]':  # Used to match only arguments, or use only stored patterns  # line 138
    return a & b if a else b  # Used to match only arguments, or use only stored patterns  # line 138

def dictUpdate(dikt: 'Dict[Any, Any]', by: 'Dict[Any, Any]') -> 'Dict[Any, Any]':  # line 140
    d = {}  # line 140
    d.update(dikt)  # line 140
    d.update(by)  # line 140
    return d  # line 140

def openIt(file: 'str', mode: 'str', compress: 'bool'=False) -> 'IO[bytes]':  # Abstraction for opening both compressed and plain files  # line 142
    return bz2.BZ2File(encode(file), mode) if compress else open(encode(file), mode + "b")  # Abstraction for opening both compressed and plain files  # line 142

def eoldet(file: 'bytes') -> '_coconut.typing.Optional[bytes]':  # line 144
    ''' Determine EOL style from a binary string. '''  # line 145
    lf = file.count(b"\n")  # type: int  # line 146
    cr = file.count(b"\r")  # type: int  # line 147
    crlf = file.count(b"\r\n")  # type: int  # line 148
    if crlf > 0:  # DOS/Windows/Symbian etc.  # line 149
        if lf != crlf or cr != crlf:  # line 150
            warn("Inconsistent CR/NL count with CR+NL. Mixed EOL style detected, may cause problems during merge")  # line 150
        return b"\r\n"  # line 151
    if lf != 0 and cr != 0:  # line 152
        warn("Inconsistent CR/NL count without CR+NL. Mixed EOL style detected, may cause problems during merge")  # line 152
    if lf > cr:  # Linux/Unix  # line 153
        return b"\n"  # Linux/Unix  # line 153
    if cr > lf:  # older 8-bit machines  # line 154
        return b"\r"  # older 8-bit machines  # line 154
    return None  # no new line contained, cannot determine  # line 155

try:  # line 157
    Splittable = TypeVar("Splittable", str, bytes)  # line 157
except:  # line 158
    pass  # line 158
def safeSplit(s: 'Splittable', d: '_coconut.typing.Optional[Splittable]'=None) -> 'List[Splittable]':  # line 159
    return s.split((("\n" if isinstance(s, str) else b"\n") if d is None else d)) if len(s) > 0 else []  # line 159

def ajoin(sep: 'str', seq: '_coconut.typing.Sequence[str]', nl: 'str'="") -> 'str':  # line 161
    return (sep + (nl + sep).join(seq)) if seq else ""  # line 161

@_coconut_tco  # line 163
def sjoin(*s: 'Tuple[Any]') -> 'str':  # line 163
    return _coconut_tail_call(" ".join, [str(e) for e in s if e != ''])  # line 163

@_coconut_tco  # line 165
def hashStr(datas: 'str') -> 'str':  # line 165
    return _coconut_tail_call(hashlib.sha256(datas.encode(UTF8)).hexdigest)  # line 165

def modified(changes: 'ChangeSet', onlyBinary: 'bool'=False) -> 'bool':  # line 167
    return len(changes.additions) > 0 or len(changes.deletions) > 0 or len(changes.modifications) > 0 or len(changes.moves) > 0  # line 167

def listindex(lizt: 'Sequence[Any]', what: 'Any', index: 'int'=0) -> 'int':  # line 169
    return lizt[index:].index(what) + index  # line 169

def getTermWidth() -> 'int':  # line 171
    try:  # line 172
        import termwidth  # line 172
    except:  # line 173
        return 80  # line 173
    return termwidth.getTermWidth()[0]  # line 174

def branchFolder(branch: 'int', base: '_coconut.typing.Optional[str]'=None, file: '_coconut.typing.Optional[str]'=None) -> 'str':  # line 176
    return os.path.join((os.getcwd() if base is None else base), metaFolder, "b%d" % branch) + ((os.sep + file) if file else "")  # line 176

def revisionFolder(branch: 'int', revision: 'int', base: '_coconut.typing.Optional[str]'=None, file: '_coconut.typing.Optional[str]'=None) -> 'str':  # line 178
    return os.path.join(branchFolder(branch, base), "r%d" % revision) + ((os.sep + file) if file else "")  # line 178

def Exit(message: 'str'="", code=1):  # line 180
    printe("[EXIT%s]" % (" %.1fs" % (time.time() - START_TIME) if verbose else "") + (" " + message + "." if message != "" else ""))  # line 180
    sys.exit(code)  # line 180

def exception(E):  # line 182
    ''' Report an exception. '''  # line 183
    printo(str(E))  # line 184
    import traceback  # line 185
    traceback.print_exc()  # line 186
    traceback.print_stack()  # line 187

def hashFile(path: 'str', compress: 'bool', saveTo: '_coconut.typing.Optional[str]'=None, callback: 'Optional[_coconut.typing.Callable[[str], None]]'=None) -> 'Tuple[str, int]':  # line 189
    ''' Calculate hash of file contents, and return compressed sized, if in write mode, or zero. '''  # line 190
    indicator = ProgressIndicator(callback) if callback else None  # type: _coconut.typing.Optional[ProgressIndicator]  # line 191
    _hash = hashlib.sha256()  # line 192
    wsize = 0  # type: int  # line 193
    if saveTo and os.path.exists(encode(saveTo)):  # line 194
        Exit("Hash conflict. Leaving revision in inconsistent state. This should happen only once in a lifetime.")  # line 194
    to = openIt(saveTo, "w", compress) if saveTo else None  # line 195
    with open(encode(path), "rb") as fd:  # line 196
        while True:  # line 197
            buffer = fd.read(bufSize)  # type: bytes  # line 198
            _hash.update(buffer)  # line 199
            if to:  # line 200
                to.write(buffer)  # line 200
            if len(buffer) < bufSize:  # line 201
                break  # line 201
            if indicator:  # line 202
                indicator.getIndicator()  # line 202
        if to:  # line 203
            to.close()  # line 204
            wsize = os.stat(encode(saveTo)).st_size  # line 205
    return (_hash.hexdigest(), wsize)  # line 206

def getAnyOfMap(map: 'Dict[str, Any]', params: '_coconut.typing.Sequence[str]', default: 'Any'=None) -> 'Any':  # line 208
    ''' Utility to find any entries of a dictionary in a list to return the dictionaries value. '''  # line 209
    for k, v in map.items():  # line 210
        if k in params:  # line 211
            return v  # line 211
    return default  # line 212

@_coconut_tco  # line 214
def strftime(timestamp: '_coconut.typing.Optional[int]'=None) -> 'str':  # line 214
    return _coconut_tail_call(time.strftime, "%Y-%m-%d %H:%M:%S", time.localtime(timestamp / 1000. if timestamp is not None else None))  # line 214

def detectAndLoad(filename: '_coconut.typing.Optional[str]'=None, content: '_coconut.typing.Optional[bytes]'=None, ignoreWhitespace: 'bool'=False) -> 'Tuple[str, bytes, _coconut.typing.Sequence[str]]':  # line 216
    lines = []  # type: _coconut.typing.Sequence[str]  # line 217
    if filename is not None:  # line 218
        with open(encode(filename), "rb") as fd:  # line 219
            content = fd.read()  # line 219
    encoding = (lambda _coconut_none_coalesce_item: sys.getdefaultencoding() if _coconut_none_coalesce_item is None else _coconut_none_coalesce_item)(detectEncoding(content))  # type: str  # line 220
    eol = eoldet(content)  # type: _coconut.typing.Optional[bytes]  # line 221
    if filename is not None:  # line 222
        with codecs.open(encode(filename), encoding=encoding) as fd2:  # line 223
            lines = safeSplit(fd2.read(), ((b"\n" if eol is None else eol)).decode(encoding))  # line 223
    elif content is not None:  # line 224
        lines = safeSplit(content.decode(encoding), ((b"\n" if eol is None else eol)).decode(encoding))  # line 225
    else:  # line 226
        return (sys.getdefaultencoding(), b"\n", [])  # line 226
    if ignoreWhitespace:  # line 227
        lines[:] = [line.replace("\t", "  ").strip() for line in lines]  # line 227
    return (encoding, eol, lines)  # line 228

@_coconut_tco  # line 230
def dataCopy(_tipe: 'Type[DataType]', _old: 'DataType', *_args, byValue: 'bool'=False, **_kwargs) -> 'DataType':  # line 230
    ''' A better makedata() version. '''  # line 231
    r = _old._asdict()  # type: Dict[str, Any]  # line 232
    r.update({k: ([e for e in v] if byValue and isinstance(v, (list, tuple, set)) else v) for k, v in _kwargs.items()})  # copy by value if required  # line 233
    return _coconut_tail_call(makedata, _tipe, *(list(_args) + [r[field] for field in _old._fields]))  # TODO also offer copy-by-value here  # line 234

def user_block_input(output: 'List[str]'):  # line 236
    sep = input("Enter end-of-text marker (default: <empty line>: ")  # type: str  # line 237
    line = sep  # type: str  # line 237
    while True:  # line 238
        line = input("> ")  # line 239
        if line == sep:  # line 240
            break  # line 240
        output.append(line)  # writes to caller-provided list reference  # line 241

def merge(file: '_coconut.typing.Optional[bytes]'=None, into: '_coconut.typing.Optional[bytes]'=None, filename: '_coconut.typing.Optional[str]'=None, intoname: '_coconut.typing.Optional[str]'=None, mergeOperation: 'MergeOperation'=MergeOperation.BOTH, charMergeOperation: 'MergeOperation'=MergeOperation.BOTH, diffOnly: 'bool'=False, eol: 'bool'=False, ignoreWhitespace: 'bool'=False) -> 'Tuple[Union[bytes, List[MergeBlock]], _coconut.typing.Optional[bytes]]':  # line 243
    ''' Merges other binary text contents 'file' (or reads file 'filename') into current text contents 'into' (or reads file 'intoname'), returning merged result.
      For update, the other version is assumed to be the "new/added" one, while for diff, the current changes are the ones "added".
      However, change direction markers are insert ("+") for elements only in into, and remove ("-") for elements only in other file (just like the diff marks +/-)
      diffOnly returns detected change blocks only, no text merging
      eol flag will use the other file's EOL marks
      in case of replace block and INSERT strategy, the change will be added **behind** the original
  '''  # line 258
    encoding = None  # type: str  # line 259
    othr = None  # type: _coconut.typing.Sequence[str]  # line 259
    othreol = None  # type: _coconut.typing.Optional[bytes]  # line 259
    curr = None  # type: _coconut.typing.Sequence[str]  # line 259
    curreol = None  # type: _coconut.typing.Optional[bytes]  # line 259
    try:  # load files line-wise and normalize line endings (keep the one of the current file) TODO document  # line 260
        encoding, othreol, othr = detectAndLoad(filename=filename, content=file, ignoreWhitespace=ignoreWhitespace)  # line 261
        encoding, curreol, curr = detectAndLoad(filename=intoname, content=into, ignoreWhitespace=ignoreWhitespace)  # line 262
    except Exception as E:  # line 263
        Exit("Cannot merge '%s' into '%s': %r" % (filename, intoname, E))  # line 263
    if None not in [othreol, curreol] and othreol != curreol:  # line 264
        warn("Differing EOL-styles detected during merge. Using current file's style for merged output")  # line 264
    output = list(difflib.Differ().compare(othr, curr))  # type: List[str]  # from generator expression  # line 265
    blocks = []  # type: List[MergeBlock]  # merged result in blocks  # line 266
    tmp = []  # type: List[str]  # block lines  # line 267
    last = " "  # type: str  # line 268
    no = None  # type: int  # line 268
    line = None  # type: str  # line 268
    for no, line in enumerate(output + ["X"]):  # EOF marker (difflib's output will never be "X" alone)  # line 269
        if line[0] == last:  # continue filling current block, no matter what type of block it is  # line 270
            tmp.append(line[2:])  # continue filling current block, no matter what type of block it is  # line 270
            continue  # continue filling current block, no matter what type of block it is  # line 270
        if line == "X" and len(tmp) == 0:  # break if nothing left to do, otherwise perform operation for stored block  # line 271
            break  # break if nothing left to do, otherwise perform operation for stored block  # line 271
        if last == " ":  # block is same in both files  # line 272
            if len(tmp) > 0:  # avoid adding empty keep block  # line 273
                blocks.append(MergeBlock(MergeBlockType.KEEP, [line for line in tmp], line=no - len(tmp)))  # avoid adding empty keep block  # line 273
        elif last == "-":  # may be a pure deletion or part of a replacement (with next block being "+")  # line 274
            blocks.append(MergeBlock(MergeBlockType.REMOVE, [line for line in tmp], line=no - len(tmp)))  # line 275
            if len(blocks) >= 2 and blocks[-2].tipe == MergeBlockType.INSERT:  # line 276
                blocks[-2] = MergeBlock(MergeBlockType.REPLACE, blocks[-1].lines, line=no - len(tmp) - 1, replaces=blocks[-2])  # remember replaced stuff with reference to other merge block TODO why -1 necessary?  # line 277
                blocks.pop()  # line 278
        elif last == "+":  # may be insertion or replacement (with previous - block)  # line 279
            blocks.append(MergeBlock(MergeBlockType.INSERT, [line for line in tmp], line=no - len(tmp)))  # first, assume simple insertion, then check for replacement  # line 280
            if len(blocks) >= 2 and blocks[-2].tipe == MergeBlockType.REMOVE:  #  and len(blocks[-1].lines) == len(blocks[-2].lines):  # requires previous block and same number of lines TODO allow multiple intra-line merge for same-length blocks  # line 281
                blocks[-2] = MergeBlock(MergeBlockType.REPLACE, blocks[-2].lines, line=no - len(tmp) - 1, replaces=blocks[-1])  # remember replaced stuff with reference to other merge block TODO why -1 necessary?  # line 282
                blocks.pop()  # remove TOS due to merging two blocks into replace or modify  # line 283
#    elif last == "?": pass # marker for intra-line change comment -> add to block info
        last = line[0]  # line 285
        tmp[:] = [line[2:]]  # only keep current line for next block  # line 286
# TODO add code to detect block moves here
    nl = othreol if eol else ((othreol if curreol is None else curreol))  # type: bytes  # no default newline, to mark "no newline"  # line 288
    debug("Diff blocks: " + repr(blocks))  # line 289
    if diffOnly:  # line 290
        return (blocks, nl)  # line 290

# now perform merge operations depending on detected blocks
    output[:] = []  # clean list of strings  # line 293
    for block in blocks:  # line 294
        if block.tipe == MergeBlockType.KEEP:  # line 295
            output.extend(block.lines)  # line 296
        elif block.tipe == MergeBlockType.INSERT and not (mergeOperation.value & MergeOperation.REMOVE.value) or block.tipe == MergeBlockType.REMOVE and (mergeOperation.value & MergeOperation.INSERT.value):  # line 297
            output.extend(block.lines)  # line 299
        elif block.tipe == MergeBlockType.REPLACE:  # complete block replacement  # line 300
            if len(block.lines) == len(block.replaces.lines) == 1:  # one-liner  # line 301
                output.append(lineMerge(block.lines[0], block.replaces.lines[0], mergeOperation=charMergeOperation))  # line 302
            elif mergeOperation == MergeOperation.ASK:  # more than one line: needs user input  # line 303
                printo(ajoin("- ", block.replaces.lines, nl="\n"))  # TODO check +/- in update mode, could be swapped  # line 304
                printo(ajoin("+ ", block.lines, nl="\n"))  # line 305
                while True:  # line 306
                    op = input(" Line replacement: *M[I]ne (+), [T]heirs (-), [B]oth, [U]ser input: ").strip().lower()  # type: str  # line 307
                    if op in "tb":  # line 308
                        output.extend(block.lines)  # line 308
                        break  # line 308
                    if op in "ib":  # line 309
                        output.extend(block.replaces.lines)  # line 309
                        break  # line 309
                    if op == "m":  # line 310
                        user_block_input(output)  # line 310
                        break  # line 310
            else:  # more than one line and not ask  # line 311
                if mergeOperation == MergeOperation.REMOVE:  # line 312
                    pass  # line 312
                elif mergeOperation == MergeOperation.BOTH:  # line 313
                    output.extend(block.lines)  # line 313
                elif mergeOperation == MergeOperation.INSERT:  # TODO optionally allow insertion BEFORE or AFTER original (order of these both lines)  # line 314
                    output.extend(list(block.replaces.lines) + list(block.lines))  # TODO optionally allow insertion BEFORE or AFTER original (order of these both lines)  # line 314
#  debug("Merge output: " + "; ".join(output))
    return (((b"\n" if nl is None else nl)).join([line.encode(encoding) for line in output]), nl)  # returning bytes  # line 316
# TODO handle check for more/less lines in found -/+ blocks to find common section and splitting prefix/suffix out

@_coconut_tco  # line 319
def lineMerge(othr: 'str', into: 'str', mergeOperation: 'MergeOperation'=MergeOperation.BOTH, diffOnly: 'bool'=False) -> 'Union[str, List[MergeBlock]]':  # line 319
    ''' Merges string 'othr' into current string 'into'.
      change direction mark is insert for elements only in into, and remove for elements only in file (according to diff marks +/-)
  '''  # line 322
    out = list(difflib.Differ().compare(othr, into))  # type: List[str]  # line 323
    blocks = []  # type: List[MergeBlock]  # line 324
    for i, line in enumerate(out):  # line 325
        if line[0] == "+":  # line 326
            if i + 1 < len(out) and out[i + 1][0] == "+":  # block will continue  # line 327
                if i > 0 and blocks[-1].tipe == MergeBlockType.INSERT:  # middle of + block  # line 328
                    blocks[-1].lines.append(line[2])  # add one more character to the accumulating list  # line 329
                else:  # first + in block  # line 330
                    blocks.append(MergeBlock(MergeBlockType.INSERT, [line[2]], i))  # line 331
            else:  # last line of + block  # line 332
                if i > 0 and blocks[-1].tipe == MergeBlockType.INSERT:  # end of a block  # line 333
                    blocks[-1].lines.append(line[2])  # line 334
                else:  # single line  # line 335
                    blocks.append(MergeBlock(MergeBlockType.INSERT, [line[2]], i))  # line 336
                if i >= 1 and blocks[-2].tipe == MergeBlockType.REMOVE:  # previous - and now last in + block creates a replacement block  # line 337
                    blocks[-2] = MergeBlock(MergeBlockType.REPLACE, blocks[-2].lines, i, replaces=blocks[-1])  # line 338
                    blocks.pop()  # line 338
        elif line[0] == "-":  # line 339
            if i > 0 and blocks[-1].tipe == MergeBlockType.REMOVE:  # part of - block  # line 340
                blocks[-1].lines.append(line[2])  # line 341
            else:  # first in block  # line 342
                blocks.append(MergeBlock(MergeBlockType.REMOVE, [line[2]], i))  # line 343
        elif line[0] == " ":  # line 344
            if i > 0 and blocks[-1].tipe == MergeBlockType.KEEP:  # part of block  # line 345
                blocks[-1].lines.append(line[2])  # line 346
            else:  # first in block  # line 347
                blocks.append(MergeBlock(MergeBlockType.KEEP, [line[2]], i))  # line 348
        else:  # line 349
            raise Exception("Cannot parse diff line %r" % line)  # line 349
    blocks[:] = [dataCopy(MergeBlock, block, lines=["".join(block.lines)], replaces=dataCopy(MergeBlock, block.replaces, lines=["".join(block.replaces.lines)]) if block.replaces else None) for block in blocks]  # line 350
    if diffOnly:  # line 351
        return blocks  # line 351
    out[:] = []  # line 352
    for i, block in enumerate(blocks):  # line 353
        if block.tipe == MergeBlockType.KEEP:  # line 354
            out.extend(block.lines)  # line 354
        elif block.tipe == MergeBlockType.REPLACE:  # line 355
            if mergeOperation == MergeOperation.ASK:  # line 356
                printo(ajoin("- ", othr))  # line 357
                printo("- " + (" " * i) + block.replaces.lines[0])  # line 358
                printo("+ " + (" " * i) + block.lines[0])  # line 359
                printo(ajoin("+ ", into))  # line 360
                while True:  # line 361
                    op = input(" Character replacement: *M[I]ne (+), [T]heirs (-), [B]oth, [U]ser input: ").strip().lower()  # type: str  # line 362
                    if op in "tb":  # line 363
                        out.extend(block.lines)  # line 363
                        break  # line 363
                    if op in "ib":  # line 364
                        out.extend(block.replaces.lines)  # line 364
                        break  # line 364
                    if op == "m":  # line 365
                        user_block_input(out)  # line 365
                        break  # line 365
            else:  # non-interactive  # line 366
                if mergeOperation == MergeOperation.REMOVE:  # line 367
                    pass  # line 367
                elif mergeOperation == MergeOperation.BOTH:  # line 368
                    out.extend(block.lines)  # line 368
                elif mergeOperation == MergeOperation.INSERT:  # line 369
                    out.extend(list(block.replaces.lines) + list(block.lines))  # line 369
        elif block.tipe == MergeBlockType.INSERT and not (mergeOperation.value & MergeOperation.REMOVE.value):  # line 370
            out.extend(block.lines)  # line 370
        elif block.tipe == MergeBlockType.REMOVE and mergeOperation.value & MergeOperation.INSERT.value:  # line 371
            out.extend(block.lines)  # line 371
# TODO ask for insert or remove as well
    return _coconut_tail_call("".join, out)  # line 373

def findSosVcsBase() -> 'Tuple[_coconut.typing.Optional[str], _coconut.typing.Optional[str], _coconut.typing.Optional[str]]':  # line 375
    ''' Attempts to find sos and legacy VCS base folders.
      Returns (SOS-repo root, VCS-repo root, VCS command)
  '''  # line 378
    debug("Detecting root folders...")  # line 379
    path = os.getcwd()  # type: str  # start in current folder, check parent until found or stopped  # line 380
    vcs = (None, None)  # type: Tuple[_coconut.typing.Optional[str], _coconut.typing.Optional[str]]  # line 381
    while not os.path.exists(encode(os.path.join(path, metaFolder))):  # line 382
        contents = set(os.listdir(path))  # type: Set[str]  # line 383
        vcss = [executable for folder, executable in vcsFolders.items() if folder in contents]  # type: _coconut.typing.Sequence[str]  # determine VCS type from dot folder  # line 384
        choice = None  # type: _coconut.typing.Optional[str]  # line 385
        if len(vcss) > 1:  # line 386
            choice = SVN if SVN in vcss else vcss[0]  # line 387
            warn("Detected more than one parallel VCS checkouts %r. Falling back to '%s'" % (vcss, choice))  # line 388
        elif len(vcss) > 0:  # line 389
            choice = vcss[0]  # line 389
        if not vcs[0] and choice:  # memorize current repo root  # line 390
            vcs = (path, choice)  # memorize current repo root  # line 390
        new = os.path.dirname(path)  # get parent path  # line 391
        if new == path:  # avoid infinite loop  # line 392
            break  # avoid infinite loop  # line 392
        path = new  # line 393
    if os.path.exists(encode(os.path.join(path, metaFolder))):  # found something  # line 394
        if vcs[0]:  # already detected vcs base and command  # line 395
            return (path, vcs[0], vcs[1])  # already detected vcs base and command  # line 395
        sos = path  # line 396
        while True:  # continue search for VCS base  # line 397
            new = os.path.dirname(path)  # get parent path  # line 398
            if new == path:  # no VCS folder found  # line 399
                return (sos, None, None)  # no VCS folder found  # line 399
            path = new  # line 400
            contents = set(os.listdir(path))  # line 401
            vcss = [executable for folder, executable in vcsFolders.items() if folder in contents]  # determine VCS type  # line 402
            choice = None  # line 403
            if len(vcss) > 1:  # line 404
                choice = SVN if SVN in vcss else vcss[0]  # line 405
                warn("Detected more than one parallel VCS checkouts %r. Falling back to '%s'" % (vcss, choice))  # line 406
            elif len(vcss) > 0:  # line 407
                choice = vcss[0]  # line 407
            if choice:  # line 408
                return (sos, path, choice)  # line 408
    return (None, vcs[0], vcs[1])  # line 409

def tokenizeGlobPattern(pattern: 'str') -> 'List[GlobBlock]':  # line 411
    index = 0  # type: int  # line 412
    out = []  # type: List[GlobBlock]  # literal = True, first index  # line 413
    while index < len(pattern):  # line 414
        if pattern[index:index + 3] in ("[?]", "[*]", "[[]", "[]]"):  # line 415
            out.append(GlobBlock(False, pattern[index:index + 3], index))  # line 415
            continue  # line 415
        if pattern[index] in "*?":  # line 416
            count = 1  # type: int  # line 417
            while index + count < len(pattern) and pattern[index] == "?" and pattern[index + count] == "?":  # line 418
                count += 1  # line 418
            out.append(GlobBlock(False, pattern[index:index + count], index))  # line 419
            index += count  # line 419
            continue  # line 419
        if pattern[index:index + 2] == "[!":  # line 420
            out.append(GlobBlock(False, pattern[index:pattern.index("]", index + 2) + 1], index))  # line 420
            index += len(out[-1][1])  # line 420
            continue  # line 420
        count = 1  # line 421
        while index + count < len(pattern) and pattern[index + count] not in "*?[":  # line 422
            count += 1  # line 422
        out.append(GlobBlock(True, pattern[index:index + count], index))  # line 423
        index += count  # line 423
    return out  # line 424

def tokenizeGlobPatterns(oldPattern: 'str', newPattern: 'str') -> 'Tuple[_coconut.typing.Sequence[GlobBlock], _coconut.typing.Sequence[GlobBlock]]':  # line 426
    ot = tokenizeGlobPattern(oldPattern)  # type: List[GlobBlock]  # line 427
    nt = tokenizeGlobPattern(newPattern)  # type: List[GlobBlock]  # line 428
#  if len(ot) != len(nt): Exit("Source and target patterns can't be translated due to differing number of parsed glob markers and literal strings")
    if len([o for o in ot if not o.isLiteral]) < len([n for n in nt if not n.isLiteral]):  # line 430
        Exit("Source and target file patterns contain differing number of glob markers and can't be translated")  # line 430
    if any((O.content != N.content for O, N in zip([o for o in ot if not o.isLiteral], [n for n in nt if not n.isLiteral]))):  # line 431
        Exit("Source and target file patterns differ in semantics")  # line 431
    return (ot, nt)  # line 432

def convertGlobFiles(filenames: '_coconut.typing.Sequence[str]', oldPattern: '_coconut.typing.Sequence[GlobBlock]', newPattern: '_coconut.typing.Sequence[GlobBlock]') -> '_coconut.typing.Sequence[Tuple[str, str]]':  # line 434
    ''' Converts given filename according to specified file patterns. No support for adjacent glob markers currently. '''  # line 435
    pairs = []  # type: List[Tuple[str, str]]  # line 436
    for filename in filenames:  # line 437
        literals = [l for l in oldPattern if l.isLiteral]  # type: List[GlobBlock]  # source literals  # line 438
        nextliteral = 0  # type: int  # line 439
        parsedOld = []  # type: List[GlobBlock2]  # line 440
        index = 0  # type: int  # line 441
        for part in oldPattern:  # match everything in the old filename  # line 442
            if part.isLiteral:  # line 443
                parsedOld.append(GlobBlock2(True, part.content, part.content))  # line 443
                index += len(part.content)  # line 443
                nextliteral += 1  # line 443
            elif part.content.startswith("?"):  # line 444
                parsedOld.append(GlobBlock2(False, part.content, filename[index:index + len(part.content)]))  # line 444
                index += len(part.content)  # line 444
            elif part.content.startswith("["):  # line 445
                parsedOld.append(GlobBlock2(False, part.content, filename[index]))  # line 445
                index += 1  # line 445
            elif part.content == "*":  # line 446
                if nextliteral >= len(literals):  # line 447
                    parsedOld.append(GlobBlock2(False, part.content, filename[index:]))  # line 447
                    break  # line 447
                nxt = filename.index(literals[nextliteral].content, index)  # type: int  # also matches empty string  # line 448
                parsedOld.append(GlobBlock2(False, part.content, filename[index:nxt]))  # line 449
                index = nxt  # line 449
            else:  # line 450
                Exit("Invalid file pattern specified for move/rename")  # line 450
        globs = [g for g in parsedOld if not g.isLiteral]  # type: List[GlobBlock2]  # line 451
        literals = [l for l in newPattern if l.isLiteral]  # target literals  # line 452
        nextliteral = 0  # line 453
        nextglob = 0  # type: int  # line 453
        outname = []  # type: List[str]  # line 454
        for part in newPattern:  # generate new filename  # line 455
            if part.isLiteral:  # line 456
                outname.append(literals[nextliteral].content)  # line 456
                nextliteral += 1  # line 456
            else:  # line 457
                outname.append(globs[nextglob].matches)  # line 457
                nextglob += 1  # line 457
        pairs.append((filename, "".join(outname)))  # line 458
    return pairs  # line 459

@_coconut_tco  # line 461
def reorderRenameActions(actions: '_coconut.typing.Sequence[Tuple[str, str]]', exitOnConflict: 'bool'=True) -> '_coconut.typing.Sequence[Tuple[str, str]]':  # line 461
    ''' Attempt to put all rename actions into an order that avoids target == source names.
      Note, that it's currently not really possible to specify patterns that make this work (swapping "*" elements with a reference).
      An alternative would be to always have one (or all) files renamed to a temporary name before renaming to target filename.
  '''  # line 465
    if not actions:  # line 466
        return []  # line 466
    sources = None  # type: List[str]  # line 467
    targets = None  # type: List[str]  # line 467
    sources, targets = [list(l) for l in zip(*actions)]  # line 468
    last = len(actions)  # type: int  # line 469
    while last > 1:  # line 470
        clean = True  # type: bool  # line 471
        for i in range(1, last):  # line 472
            try:  # line 473
                index = targets[:i].index(sources[i])  # type: int  # line 474
                sources.insert(index, sources.pop(i))  # bubble up the action right before conflict  # line 475
                targets.insert(index, targets.pop(i))  # line 476
                clean = False  # line 477
            except:  # target not found in sources: good!  # line 478
                continue  # target not found in sources: good!  # line 478
        if clean:  # line 479
            break  # line 479
        last -= 1  # we know that the last entry in the list has the least conflicts, so we can disregard it in the next iteration  # line 480
    if exitOnConflict:  # line 481
        for i in range(1, len(actions)):  # line 482
            if sources[i] in targets[:i]:  # line 483
                Exit("There is no order of renaming actions that avoids copying over not-yet renamed files: '%s' is contained in matching source filenames" % (targets[i]))  # line 483
    return _coconut_tail_call(list, zip(sources, targets))  # convert to list to avoid generators  # line 484

def relativize(root: 'str', filepath: 'str') -> 'Tuple[str, str]':  # line 486
    ''' Determine OS-independent relative folder path, and relative pattern path. '''  # line 487
    relpath = os.path.relpath(os.path.dirname(os.path.abspath(filepath)), root).replace(os.sep, SLASH)  # line 488
    return relpath, os.path.join(relpath, os.path.basename(filepath)).replace(os.sep, SLASH)  # line 489

def parseOnlyOptions(root: 'str', options: '_coconut.typing.Sequence[str]') -> 'Tuple[_coconut.typing.Optional[FrozenSet[str]], _coconut.typing.Optional[FrozenSet[str]]]':  # line 491
    ''' Returns set of --only arguments, and set or --except arguments. '''  # line 492
    cwd = os.getcwd()  # type: str  # line 493
    onlys = []  # type: List[str]  # line 494
    excps = []  # type: List[str]  # line 494
    index = 0  # type: int  # line 494
    while True:  # line 495
        try:  # line 496
            index = 1 + listindex(options, "--only", index)  # line 497
            onlys.append(options[index])  # line 498
        except:  # line 499
            break  # line 499
    index = 0  # line 500
    while True:  # line 501
        try:  # line 502
            index = 1 + listindex(options, "--except", index)  # line 503
            excps.append(options[index])  # line 504
        except:  # line 505
            break  # line 505
    return (frozenset((relativize(root, o)[1] for o in onlys)) if onlys else None, frozenset((relativize(root, e)[1] for e in excps)) if excps else None)  # line 506
